%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$

-module(erl.lang.term).

-export([arith_equal/2, equal/2, external_size/1, greater/2,
	 greater_or_equal/2, hash/1, hash/2, is_atom/1, is_binary/1,
	 is_constant/1, is_float/1, is_function/1, is_integer/1,
	 is_list/1, is_number/1, is_pid/1, is_port/1, is_record/3,
	 is_reference/1, is_tuple/1, less/2, less_or_equal/2, max/2,
	 min/2, not_arith_equal/2, not_equal/2, size/1, to_string/1]).


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X == Y</code> (arithmetic equality).
%% @see equal/2
%% @see not_arith_equal/2

arith_equal(X, Y) -> X == Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X =:= Y</code> (exact equality).
%% @see arith_equal/2
%% @see not_equal/2

equal(X, Y) -> X =:= Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X /= Y</code> (arithmetic inequality).
%% @see not_equal/2
%% @see arith_equal/2

not_arith_equal(X, Y) -> X /= Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X =/= Y</code> (exact inequality).
%% @see not_arith_equal/2
%% @see equal/2

not_equal(X, Y) -> X =/= Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X &gt; Y</code>.
%% @see greater_or_equal/2
%% @see less/2

greater(X, Y) -> X > Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X &gt;= Y</code>.
%% @see greater/2
%% @see equal/2

greater_or_equal(X, Y) -> X >= Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X &lt; Y</code>.
%% @see less_or_equal/2
%% @see greater/2

less(X, Y) -> X < Y.


%% @spec (X::term(), Y::term()) -> bool()
%% @doc Returns the result of <code>X =&lt; Y</code>.
%% @see less/2

less_or_equal(X, Y) -> X =< Y.


%% @spec (term()) -> bool()

is_atom(Term) ->
    erl.lang.term:is_atom(Term).	% Built-in function


%% @spec (term()) -> bool()

is_binary(Term) ->
    erl.lang.term:is_binary(Term).	% Built-in function


%% @spec (term()) -> bool()

is_constant(Term) ->
    erl.lang.term:is_constant(Term).	% Built-in function


%% @spec (term()) -> bool()

is_float(Term) ->
    erl.lang.term:is_float(Term).	% Built-in function


%% @spec (term()) -> bool()

is_function(Term) ->
    erl.lang.term:is_function(Term).	% Built-in function


%% @spec (term()) -> bool()

is_integer(Term) ->
    erl.lang.term:is_integer(Term).	% Built-in function


%% @spec (term()) -> bool()

is_number(Term) ->
    erl.lang.term:is_number(Term).	% Built-in function


%% @spec (term()) -> bool()

is_pid(Term) ->
    erl.lang.term:is_pid(Term).		% Built-in function


%% @spec (term()) -> bool()

is_port(Term) ->
    erl.lang.term:is_port(Term).	% Built-in function


%% @spec (term()) -> bool()

is_reference(Term) ->
    erl.lang.term:is_reference(Term).	% Built-in function


%% @spec (term()) -> bool()

is_tuple(Term) ->
    erl.lang.term:is_tuple(Term).	% Built-in function


%% @spec (term()) -> bool()

is_list([_ | _]) -> true;
is_list([]) -> true;
is_list(_) -> false.


%% @spec (Term::term(), Name::atom(), Size::integer()) -> bool()
%% @doc Avoid using this function! It exists for internal purposes only.

is_record(Term, Name, Size)
  when is_atom(Name), is_integer(Size), Size > 0 ->
    if is_tuple(Term), Size =:= size(Term), Name =:= element(1, Term) ->
	    true;
       true ->
	    false
    end;
is_record(Term, Name, Size) ->
    erl.lang:error(badarg, [Term, Name, Size]).


%% @spec (term()) -> integer()
%% @doc Returns the number of elements of a tuple, or the number of
%% bytes in a binary.

size(Term) ->
    erl.lang.term:size(Term).		% Built-in function


%% @spec (Term::term()) -> integer()
%% @doc Portable hash function. Equivalent to <code>hash(Term,
%% Range)</code> with a default <code>Range</code> of <code>2^27</code>.

hash(Term) ->
    erl.lang.term:hash(Term).		% Built-in function


%% @spec (Term::term(), Range::integer()) -> integer()
%%
%% @doc Portable hash function. Will give the same value for the same
%% Erlang term regardless of machine architecture and ERTS version
%% <code>Range</code> may be between <code>1</code> and
%% <code>2^32</code>. The function returns a hash value for
%% <code>Term</code> within the range <code>0</code> and
%% <code>Range-1</code>.

hash(Term, Range) ->
    erl.lang.term:hash(Term, Range).	% Built-in function


%% @spec (term()) -> integer()
%% @doc Not officially documented.

external_size(Term) ->
    erl.lang.term:external_size(Term).	% Built-in function


%% @spec (term(), term()) -> term()
%% @doc Returns the minimum of two elements, in the Erlang term order.

min(X, Y) when X =< Y -> X;
min(_, Y) -> Y.


%% @spec (term(), term()) -> term()
%% @doc Returns the maximum of two elements, in the Erlang term order.

max(X, Y) when X >= Y -> X;
max(_, Y) -> Y.


%% @spec (term()) -> string()
%%
%% @doc Returns a printable string representation of the given term. The
%% string is not necessarily unique for the term, and does not contain
%% line breaks or other visual formatting; however, lists of characters
%% are printed as string literals (within double-quotes) when possible.
%% The main purpose is for debugging and similar inspection of terms.

%% Note: vectors are not officially supported.

to_string([]) -> "[]";
to_string(X = [_ | _]) ->
    case erl.lang.string:is_printable(X) of
	true -> erl.lang.string:literal(X);
	false -> "[" ++ list_to_string(X)
    end;
to_string(X) when is_atom(X) -> erl.lang.atom:to_string(X);
to_string(X) when is_binary(X) -> erl.lang.binary:to_list(X);
to_string(X) when is_float(X) ->
    erl.lang.number:tidy(erl.lang.float:to_string(X));
to_string(X) when is_function(X) -> erl.lang.function:to_string(X);
to_string(X) when is_integer(X) -> erl.lang.integer:to_string(X);
to_string(X) when is_pid(X) -> erl.lang.proc:pid_to_string(X);
to_string(X) when is_port(X) -> erl.lang.port:to_string(X);
to_string(X) when is_reference(X) -> erl.lang.ref:to_string(X);
to_string(X) when is_tuple(X) ->
    "{" ++ seq_to_string(erl.lang.tuple:to_list(X), "}");
to_string(X) when size(X) >= 0 ->   % must be vector if guard succeeds
    "#Vector<" ++ seq_to_string(erl.lang.vector:to_list(X), ">").

seq_to_string([H], Tail) ->
    to_string(H) ++ Tail;
seq_to_string([H | T], Tail) ->
    to_string(H) ++ "," ++ seq_to_string(T, Tail);
seq_to_string([], Tail) -> Tail.

list_to_string([X]) ->
    to_string(X) ++ "]";
list_to_string([H | T]) when is_list(T) ->
    to_string(H) ++ "," ++ list_to_string(T);
list_to_string([H | T]) ->
    to_string(H) ++ "|" ++ to_string(T) ++ "]";
list_to_string([]) -> "".
