%%%-------------------------------------------------------------------
%%% File    : hipe_icode_fp.erl
%%% Author  : Tobias Lindahl <tobiasl@it.uu.se>
%%% Description : One pass analysis to find floating point values. 
%%%               Mapping to fp variables and creation of fp ebbs.
%%%
%%% Created : 23 Apr 2003 by Tobias Lindahl <tobiasl@it.uu.se>
%%%
%%% CVS      :
%%%              $Author: kostis $
%%%              $Date: 2003/11/08 15:16:28 $
%%%              $Revision: 1.9 $

%%%-------------------------------------------------------------------
-module(hipe_icode_fp).

-export([cfg/1,
	 cfg/2]).

-export([print_state_info/1]).

-import(erl_types, [
		    t_any/0,
		    t_float/0,
		    t_from_term/1,
		    t_inf/2,
		    t_is_any/1,
		    t_is_float/1,
		    t_sup/2,
		    t_to_string/1,
		    t_undefined/0]).

-record(state, {info, info_map, block_map, cfg}).

cfg(Cfg) ->  
  %%hipe_icode_cfg:pp(Cfg),  
  State = analyse(Cfg),
  %%print_state_info(State),
  NewState = place_fp_blocks(State),
  NewCfg = state__cfg(NewState),
  %%hipe_icode_cfg:pp(NewCfg),
  NewCfg.

cfg(Cfg, InfoMap) ->
  %%hipe_icode_cfg:pp(Cfg),
  State = new_state(Cfg, InfoMap),
  NewState = place_fp_blocks(State),
  NewCfg = state__cfg(NewState),
  %%hipe_icode_cfg:pp(NewCfg),
  NewCfg.


%%____________________________________________________________
%%
%% Single pass analysis that only focus on floats.
%%

analyse(Cfg)->
  State = new_state(Cfg),
  Labels = hipe_icode_cfg:reverse_postorder(Cfg),
  analyse_blocks(Labels, State).

analyse_blocks(Labels, State)->
  analyse_blocks(Labels, Labels, State).

analyse_blocks(Labels, [Label|Left], State)->
  Info = state__info_in(State, Label),
  NewState = analyse_block(Label, Info, State),
  analyse_blocks(Labels, Left, NewState);
analyse_blocks(_Labels, [], State) ->
  State.

analyse_block(Label, InfoIn, State)->
  %%io:format("Handling ~w\n", [Label]),
  BB = state__bb(State, Label),
  Code = hipe_bb:code(BB),
  Last = hipe_bb:last(BB),
  NewInfoIn = analyse_insns(Code, InfoIn),
  NewState = state__info_out_update(State, Label, NewInfoIn),

  case hipe_icode:type(Last) of
    type ->
      UpdateInfo = do_type(Last, NewInfoIn),
      %%io:format("Update info for ~w:\n~w", [Label, UpdateInfo]),
      do_updates(NewState, UpdateInfo);
    _ ->
      UpdateInfo = [{X, NewInfoIn}||X<-state__succ(NewState, Label)],
      %%io:format("Update info for ~w:\n~w", [Label, UpdateInfo]),
      do_updates(NewState, UpdateInfo)
  end.

analyse_insns([I|Insns], Info)->
  NewInfo = 
    case hipe_icode:type(I) of
      mov ->
	do_mov(I, Info);
      call ->
	do_call(I, Info);
      phi ->
	Type = join_list(args(I), Info),
	enter_defines(I, Type, Info);
      _ ->
	enter_defines(I, t_any(), Info)
    end,
  analyse_insns(Insns, NewInfo);  
analyse_insns([], Info) ->
  Info.

do_mov(I, Info)->
  %% Can't use uses/1 since we must keep constants.
  Src = hipe_icode:mov_src(I), 
  case hipe_icode:is_const(Src) of
    false ->
      %% Make the destination point to the source.
      enter_defines(I, Src, Info);
    true ->
      ConstType = t_from_term(hipe_icode:const_value(Src)),
      enter_defines(I, ConstType, Info)
  end.

do_call(I, Info)->
  case hipe_icode:call_type(I) of
    remote ->
      case hipe_icode:call_fun(I) of
	{M, F, A} ->
	  ArgTypes = lookup_type_list(args(I), Info),
	  Type = erl_bif_types:type(M, F, A, ArgTypes),
	  enter_defines(I, Type, Info);
	_ ->
	  Info
      end;
    local ->
      case lists:keysearch(dst_type, 1, hipe_icode:info(I)) of
	false ->
	  enter_defines(I, t_any(), Info);
	{value,{_, [Type]}}->
	  enter_defines(I, Type, Info)
      end;
    primop ->
      Fun = hipe_icode:call_fun(I),
      ArgType = lookup_type_list(args(I), Info),
      DstType = hipe_rtl_primops:type(Fun, ArgType),
      enter_defines(I, DstType, Info)
  end.

do_type(I, Info)->
  [Var] = args(I),
  VarInfo = lookup_type(I, Info),
  TrueLab = hipe_icode:type_true_label(I),
  FalseLab = hipe_icode:type_false_label(I),
  
  case hipe_icode:type_type(I) of
    float ->
      TrueType = t_inf(t_float(), VarInfo),
      TrueInfo = enter_type(Var, TrueType, Info),
      FalseInfo = enter_type(Var, t_any(), Info),
      [{TrueLab, TrueInfo}, {FalseLab, FalseInfo}];
    _ ->
      [{TrueLab, Info}, {FalseLab, Info}]
  end.

do_updates(State, [{Label, Info}|Tail])->
  case state__info_in_update(State, Label, Info) of
    fixpoint ->
      do_updates(State, Tail);
    NewState ->
      do_updates(NewState, Tail)
  end;
do_updates(State, []) ->
  State.

%%____________________________________________________________
%%
%% Make float ebbs
%%

place_fp_blocks(State)->
  WorkList = new_worklist(State),
  transform_block(WorkList, State).

transform_block(WorkList, State)->
  case get_work(WorkList) of
    none ->
      finalize(State);
    {Label, NewWorkList} ->
      {Prelude, Map} = do_prelude(State, Label),
      {NewState, InBlock} = state__in_block(State, Label),
      BB = state__bb(State, Label),
      Code = hipe_bb:code(BB),
      Info = state__info_out(State, Label),
      {NewMap, NewInBlock, NewCode1} = 
	transform_instrs(Code, Map, init_map(Map), Info, InBlock, []),
      NewCode2 = handle_unchecked_end(state__succ(NewState, Label), 
				      NewCode1, NewInBlock),
      NewBB = hipe_bb:code_update(BB, Prelude++NewCode2),
      NewState0 = state__bb_update(NewState, Label, NewBB),
      case state__map_update(NewState0, Label, NewMap, NewInBlock) of
	fixpoint ->
	  transform_block(NewWorkList, NewState0);
	NewState1 ->
	  Succ = state__succ(NewState1, Label),
	  NewWorkList1 = add_work(NewWorkList, Succ),
	  transform_block(NewWorkList1, NewState1)
      end
  end.


transform_instrs([I|Left], PhiMap, Map, Info, InBlock, Acc)->
  case hipe_icode:type(I) of
    phi ->
      Uses = args(I),
      case [Y||X<-Uses,(Y=lookup(X, PhiMap))/=none] of
	[] ->
	  %% No ordinary variables from the argument has been untagged.
	  transform_instrs(Left, PhiMap, Map, Info, InBlock, [I|Acc]);
	_ ->
	  %% We might have to add a new phi instruction to handle that
	  %% at least one of the ordinary variables in the argument
	  %% has been untagged.
	  Defines = defines(I),
	  case hipe_icode:is_fvar(hd(Defines)) of
	    true ->
	      %% This was already a fp phi instruction.
	      transform_instrs(Left, PhiMap, Map, Info, 
			       InBlock, [I|Acc]);
	    false ->
	      NewMap = add_new_bindings_assigned(Defines, Map),
	      Dst = lookup(hd(Defines), NewMap),
	      NewI = subst_phi(I, Dst, PhiMap),
	      transform_instrs(Left, PhiMap, NewMap, Info, 
			       InBlock, [NewI, I|Acc])
	  end
      end;
    call ->
      %% TODO: Handle overlapping fp ebbs
      case hipe_icode:call_fun(I) of
	fcheckerror ->
	  case InBlock of
	    {true, _} ->
	      transform_instrs(Left, PhiMap, Map, Info, false, [I|Acc]);
	    false -> 
	      exit('fcheckerror does not end fp ebb')
	  end;
	fclearerror ->
	  case InBlock of
	    false -> 
	      transform_instrs(Left, PhiMap, Map, Info, {true, []}, [I|Acc]);
	    {true, _} -> 
	      transform_instrs(Left, PhiMap, Map, Info, InBlock, Acc)
	  end;
	X when X == unsafe_untag_float; X == conv_to_float ->
	  [Dst] = defines(I),
	  case args(I) of
	    [] -> %% Constant
	      transform_instrs(Left, PhiMap, Map, Info, InBlock, [I|Acc]);
	    [Src] ->
	      case lookup(Src, Map) of
		none ->
		  NewMap = gb_trees:enter(Src, {assigned, Dst}, Map),
		  transform_instrs(Left, PhiMap, NewMap, Info, 
				   InBlock, [I|Acc]);
		Dst ->
		  %% This is the instruction that untagged the variable.
		  transform_instrs(Left, PhiMap, Map, Info, 
				   InBlock, [I|Acc]);
		FVar -> 
		  %% The variable was already untagged. 
		  %% This instruction can be changed to a fmove.
		  NewI = hipe_icode:mk_fmov(Dst, FVar),
		  transform_instrs(Left, PhiMap, Map, Info, 
				   InBlock, [NewI|Acc])
	      end
	  end;
	unsafe_tag_float ->
	  [Dst] = defines(I),
	  [Src] = args(I),
	  NewMap = gb_trees:enter(Dst, {assigned, Src}, Map),
	  transform_instrs(Left, PhiMap, NewMap, Info, InBlock, [I|Acc]);
	_ ->
	  {NewMap, NewInBlock, NewAcc} =
	    check_for_fop_candidates(I, Map, Info, InBlock, Acc),
	  transform_instrs(Left, PhiMap, NewMap, Info, NewInBlock, NewAcc)
      end;
    enter ->
      {NewMap, NewInBlock, NewAcc} =
	check_for_fop_candidates(I, Map, Info, InBlock, Acc),
      transform_instrs(Left, PhiMap, NewMap, Info, NewInBlock, NewAcc);
    _ ->
      {NewMap, NewInBlock, NewAcc} =
	handle_untagged_arguments(I, Map, Acc, InBlock),
      transform_instrs(Left, PhiMap, NewMap, Info, NewInBlock, NewAcc)
  end;
transform_instrs([], _PhiMap, Map, _Info, InBlock, Acc) ->
  {Map, InBlock, lists:reverse(Acc)}.

check_for_fop_candidates(I, Map, Info, InBlock, Acc)->
  case is_fop_cand(I, Info) of
    false ->
      handle_untagged_arguments(I, Map, Acc, InBlock);
    true ->
      {Op, Cont, Fail} = 
	case hipe_icode:type(I) of
	  call ->
	    Fail0 = hipe_icode:call_fail(I),
	    Cont0 = hipe_icode:call_continuation(I),
	    Op0 = fun_to_fop(hipe_icode:call_fun(I)), 
	    {Op0, Cont0, Fail0};
	  enter ->
	    {fun_to_fop(hipe_icode:enter_fun(I)), [], []}
	end,
      case Fail of
	[] ->
	  Uses = args(I),
	  Defines = defines(I),
	  Convs =  [X||X <- remove_duplicates(Uses), lookup(X, Map)==none],
	  NewMap0 = add_new_bindings_assigned(Convs, Map),
	  NewMap = add_new_bindings_unassigned(Defines, NewMap0),
	  ConvIns = get_conv_instrs(Convs, NewMap, Info),
	  NewI = hipe_icode:mk_primop(lookup_list(Defines, NewMap), Op,
				      lookup_list(Uses, NewMap), Cont, Fail),
	  case InBlock of
	    {true, Fail} -> 
	      {NewMap, InBlock, [NewI|ConvIns]++Acc};
	    {true, _NewFail} -> 
	      %% TODO: Find out if this ever happens. If so, handle it!
	      exit('Different catches');
	    false ->
	      BlockStart = hipe_icode:mk_primop([], fclearerror, []),
	      {NewMap, {true, Fail}, [NewI, BlockStart|ConvIns]++Acc}
	  end;
	_ -> %% Bailing out! Can't handle instructions in catches (yet).
	  handle_untagged_arguments(I, Map, Acc, InBlock)
      end
  end.


%% If this is an instructions that need to operate on tagged values
%% that currently are untagged, we must and tag the values and perhaps
%% end the fp ebb.

handle_untagged_arguments(I, Map, Acc, InBlock)->
  case lists:filter(fun(X)->must_be_tagged(X, Map)end, uses(I)) of
    [] ->
      case hipe_icode:type(I) of
	call ->
	  case hipe_bif:is_bif(hipe_icode:call_fun(I)) of
	    false ->
	      case InBlock of
		{true, FailLab} ->
		  Fcheck = hipe_icode:mk_primop([], fcheckerror, [],
						[], FailLab),      
		  {Map, false, [I, Fcheck|Acc]};
		false ->
		  {Map, InBlock, [I|Acc]}
	      end;
	    true -> 
	      {Map, InBlock, [I|Acc]}
	  end;
	_ -> 
	  {Map, InBlock, [I|Acc]}
      end;
    Tag ->
      TagIntrs = 
	[hipe_icode:mk_primop([Dst], unsafe_tag_float, 
			      [gb_trees:get(Dst, Map)])||
	  Dst<-Tag],
      case InBlock of
	{true, FailLab} ->
	  Fcheck = hipe_icode:mk_primop([], fcheckerror, [],
					[], FailLab),      
	  {Map, false, [I|TagIntrs]++[Fcheck|Acc]};
	false ->
	  {Map, InBlock, [I|TagIntrs]++Acc}
      end
  end.



%% Add phi nodes for untagged fp values.

do_prelude(State, Label)->  
  Map = state__map(State, Label),
  case state__pred(State, Label) of
    [_, _] ->
      Map1 = gb_trees:to_list(Map),
      lists:foldl(fun(X, Acc)->get_phi(X, Acc)end, {[], Map}, Map1);
    _ -> {[], Map}
  end.

get_phi({Var, PredList}, {InsAcc, Map})->
  case all_args_equal(PredList) of
    true ->
      {InsAcc, Map};
    false ->
      FVar = hipe_icode:mk_new_fvar(),
      NewMap = gb_trees:enter(Var, FVar, Map),
      PredList1 = strip_of_assigned(PredList, []),
      Phi = hipe_icode:mk_phi(FVar, PredList1),
      {[Phi|InsAcc], NewMap}
  end.

strip_of_assigned([{Pred, {assigned, Val}}|Left], Acc)->
  strip_of_assigned(Left, [{Pred, Val}|Acc]);
strip_of_assigned([Tuple|Left], Acc) ->
  strip_of_assigned(Left, [Tuple|Acc]);
strip_of_assigned([], Acc) ->
  Acc.


%% all_args_equal returns true if the mapping for a variable is the
%% same from all predecessors, i.e., we do not need a phi-node.

all_args_equal([{_, FVar}|Left])->
  all_args_equal(Left, FVar).

all_args_equal([{_, FVar1}|Left], FVar1)->
  all_args_equal(Left, FVar1);
all_args_equal([], _) ->
  true;
all_args_equal(_, _) ->
  false.
  

%% If the block has no successors and we still are in a fp ebb we must
%% end it to make sure we don't have any unchecked fp exceptions.

handle_unchecked_end(Succ, Code, InBlock)->
  case Succ of
    [] ->
      case InBlock of
	{true, _} ->
	  {TopCode, Last} = split_code(Code),
	  NewI = hipe_icode:mk_primop([], fcheckerror, []),
	  TopCode++[NewI, Last];
	false ->
	  Code
      end;
    _ ->
      Code
  end.

split_code(Code)->
  split_code(Code, []).

split_code([I|[]], Acc) ->
  {lists:reverse(Acc), I};
split_code([I|Left], Acc)->
  split_code(Left, [I|Acc]).


%% When all code is mapped to fp instructions we must make sure that
%% the fp ebb information goin in to each block is the same as the
%% information coming out of each predecessor. Otherwise, we must add
%% a block in between.

finalize(State)->
  Edges = needs_fclearerror(State),
  finalize(Edges, State).

finalize([{From, To}|Left], State) ->
  NewState = add_fp_ebb_fixup(From, To, fclearerror, [], State),
  finalize(Left, NewState);
finalize([], State) ->
  State.

needs_fclearerror(State = #state{cfg=Cfg}) ->
  Labels = hipe_icode_cfg:labels(Cfg),
  needs_fclearerror(Labels, State, []).

needs_fclearerror([Label|Left], State, Acc) ->
  Succ = state__succ(State, Label),
  NewEdges = needs_fclearerror0(Label, Succ, State, []),
  needs_fclearerror(Left, State, NewEdges ++ Acc);
needs_fclearerror([], _State, Acc) ->
  Acc.

needs_fclearerror0(From, [To|Left], State, Acc) ->
  FromInBlock = state__get_in_block_out(State, From),
  ToInBlock = state__get_in_block_in(State, To),
  case FromInBlock of
    ToInBlock ->
      needs_fclearerror0(From, Left, State, Acc);
    false ->
      needs_fclearerror0(From, Left, State, [{From, To}|Acc]);
    {true, _} ->%% Should not happen!
      exit('Different fp ebbs into same block')
  end;
needs_fclearerror0(_From, [], _State, Acc) ->
  Acc.

add_fp_ebb_fixup(From, To, Op, Fail, State)->
  FCCode = [hipe_icode:mk_primop([], Op, [], To, Fail)],
  FCBB = hipe_bb:mk_bb(FCCode),
  FCLabel = hipe_icode:label_name(hipe_icode:mk_new_label()),
  NewState = state__bb_update(State, FCLabel, FCBB),
  NewState1 = state__redirect(NewState, From, To, FCLabel),
  ToBB = state__bb(NewState, To),
  ToCode = hipe_bb:code(ToBB),
  NewToCode = redirect_phis(ToCode, From, FCLabel),
  NewToBB = hipe_bb:code_update(ToBB, NewToCode),
  state__bb_update(NewState1, To, NewToBB).

redirect_phis(Code, OldFrom, NewFrom)->
  redirect_phis(Code, OldFrom, NewFrom, []).

redirect_phis([I|Left], OldFrom, NewFrom, Acc)->
  case hipe_icode:type(I) of
    phi ->
      NewI = hipe_icode:phi_redirect_pred(I, OldFrom, NewFrom),
      redirect_phis(Left, OldFrom, NewFrom, [NewI|Acc]);
    _ ->
      lists:reverse(Acc)++[I|Left]
  end;
redirect_phis([], _OldFrom, _NewFrom, Acc) ->
  lists:reverse(Acc).

subst_phi(I, Dst, Map)->
  ArgList = subst_phi_uses0(hipe_icode:phi_arglist(I), Map, []),
  hipe_icode:mk_phi(Dst, ArgList).

subst_phi_uses0([{Pred, Var}|Left], Map, Acc)->
  case lookup(Var, Map) of
    [{Pred, {assigned, NewVar}}]-> 
      subst_phi_uses0(Left, Map, [{Pred, NewVar}|Acc]);
    [{Pred, NewVar}]-> 
      subst_phi_uses0(Left, Map, [{Pred, NewVar}|Acc]);
    _ -> 
      subst_phi_uses0(Left, Map, [{Pred, Var}|Acc])
  end;
subst_phi_uses0([], _Map, Acc) ->
  Acc.


%%____________________________________________________________
%%
%% Help functions
%%

%% ------------------------------------------------------------ 
%% Handling the gb_tree

empty()->
  gb_trees:empty().

lookup(Key, Tree)->
  case hipe_icode:is_const(Key) of
    %% This can be true if the same constant has been
    %% untagged more than once
    true -> none;
    false ->
      case gb_trees:lookup(Key, Tree) of
	none -> none;
	{value, {assigned, Val}} -> Val;
	{value, Val} -> Val
      end
  end.

lookup_list(List, Info)->
  lookup_list(List, fun lookup/2, Info, []).

lookup_list([H|T], Fun, Info, Acc)->
  lookup_list(T, Fun, Info, [Fun(H, Info)|Acc]);
lookup_list([], _,  _, Acc) ->
  lists:reverse(Acc).


lookup_type(Var, Tree)->
  case gb_trees:lookup(Var, Tree) of
    none ->
      case hipe_icode:is_const(Var) of
	true -> hipe_icode_type:const_type(Var);
	false -> t_undefined()
      end;
    {value, Val} ->
      case hipe_icode:is_var(Val) of
	true ->
	  lookup(Val, Tree);
	false ->
	  Val
      end
  end.

lookup_type_list(List, Info)->
  lookup_list(List, fun lookup_type/2, Info, []).


enter_type([Key], Value, Tree)->
  enter_type(Key, Value, Tree);
enter_type(Key, Value, Tree)->
  case t_is_any(Value) of
    true ->
      Tree;
    false ->
      case hipe_icode:is_const(Value) of
	true -> Tree;
	false ->enter_to_leaf(Key, Value, Tree)
      end
  end.

enter_to_leaf(Key, Value, Tree)->
  case gb_trees:lookup(Key, Tree) of
    {value, Value} ->
      Tree;
    {value, Val} ->
      case hipe_icode:is_var(Val) of
	true->
	  enter_to_leaf(Val, Value, Tree);
	false ->
	  gb_trees:enter(Key, Value, Tree)
      end;
    none ->
      gb_trees:insert(Key, Value, Tree)
  end.

enter_defines(I, Types, Info) when is_list(Types)->
  case defines(I) of
    []-> Info;
    Def->
      {NewInfo, _} =
	lists:foldl(fun(X, {Info, [Type|Tail]})->
			{enter_type(X,Type,Info), Tail}end,
		    {Info, Types}, Def),
      NewInfo
  end;
enter_defines(I, Type, Info)->
  case defines(I) of
    []-> Info;
    Def->
      lists:foldl(fun(X, Acc)->enter_type(X, Type, Acc)end, Info, Def)
  end.

join_list(List, Info)->
  join_list(List, Info, t_undefined()).
 
join_list([H|T], Info, Acc)->
  Type = t_sup(lookup_type(H, Info), Acc),
  join_list(T, Info, Type);
join_list([], _, Acc) ->
  Acc.

join_info_in([{Var, Type}|Tail], InfoIn)->
  case gb_trees:lookup(Var, InfoIn) of
    none ->
      join_info_in(Tail, enter_type(Var, Type, InfoIn));
    {value, Type} ->
      join_info_in(Tail, InfoIn);
    {value, OldType}->
      join_info_in(Tail, enter_type(Var, t_sup(OldType, Type), InfoIn))
  end;
join_info_in([], InfoIn) ->
  InfoIn.


%% ------------------------------------------------------------ 
%% Handling the map from variables to fp-variables

join_maps(Pred, BlockMap)->
%%  io:format("Joining maps for ~w\n", [Pred]),
  join_maps(Pred, BlockMap, empty()).

join_maps([Pred|Left], BlockMap, Map)->
  case gb_trees:lookup(Pred, BlockMap) of
    none ->
      join_maps(Left, BlockMap, Map);
    {value, OldMap} ->
      NewMap = join_maps0(gb_trees:to_list(OldMap), Pred, Map),
      join_maps(Left, BlockMap, NewMap)
  end;
join_maps([], _, Map) ->
%%  io:format("The resulting map is ~w\n", [gb_trees:to_list(Map)]),
  Map.

join_maps0([{Var, FVar}|Tail], Pred,  Map)->
  case gb_trees:lookup(Var, Map) of
    none ->
      join_maps0(Tail, Pred, 
		 gb_trees:enter(Var, [{Pred, FVar}], Map));
    {value, {Pred, FVar}} ->
      join_maps0(Tail, Pred, Map);
    {value, Val} ->
      join_maps0(Tail, Pred, 
		 gb_trees:enter(Var, [{Pred, FVar}|Val], Map))
  end;
join_maps0([], _, Map) ->
  Map.


%% We make a difference between values that has been assigned as
%% tagged variables and those that has got a 'virtual' binding.

add_new_bindings_unassigned([Var|Left], Map)->
  FVar = hipe_icode:mk_new_fvar(),
  add_new_bindings_unassigned(Left, gb_trees:insert(Var, FVar, Map));
add_new_bindings_unassigned([], Map) ->
  Map.

add_new_bindings_assigned([Var|Left], Map)->
  case lookup(Var, Map) of
    none ->
      FVar = hipe_icode:mk_new_fvar(),
      NewMap = gb_trees:insert(Var, {assigned, FVar}, Map),
      add_new_bindings_assigned(Left, NewMap);
    _ ->
      add_new_bindings_assigned(Left, Map)
  end;
add_new_bindings_assigned([], Map) ->
  Map.

%% The map has information about from which predecessor a particular
%% binding comes from. When this information has been used we strip
%% the map of it.

init_map(Map)->
  init_map(gb_trees:to_list(Map), empty()).

init_map([{Var, [{_, FVar}|_]}|Left], Acc)->
  init_map(Left, gb_trees:insert(Var, FVar, Acc));
init_map([{Var, FVar}|Left], Acc)->
  init_map(Left, gb_trees:insert(Var, FVar, Acc));
init_map([], Acc) ->
  Acc.

join_inblock(Pred, Map)->
  join_inblock(Pred, Map, false).

join_inblock([Pred|Left], Map, Current)->
  case gb_trees:lookup({inblock_out, Pred}, Map) of
    none ->
      join_inblock(Left, Map, Current);
    {value, Current} ->
      join_inblock(Left, Map, Current);
    {value, false} ->
      join_inblock(Left, Map, Current);
    {value, Other} ->
      case Current of
	false ->
	  join_inblock(Left, Map, Other);
	{value, _}-> %% This shouldn't happen.
	  exit('Different fp ebbs into same block')
      end
  end;
join_inblock([], _Map, Current) ->
  Current.

%% ------------------------------------------------------------ 
%% Help functions for the transformation from ordinary instruction to
%% fp-instruction

is_fop_cand(I, Info)->
  case hipe_icode:type(I) of
    call ->
      case fun_to_fop(hipe_icode:call_fun(I)) of
	false -> false;
	_ -> any_is_float(hipe_icode:call_args(I), Info)
      end;
    enter ->
      case fun_to_fop(hipe_icode:enter_fun(I)) of
	false -> false;
	_ -> any_is_float(hipe_icode:enter_args(I), Info)
      end;
    _ ->
      false
  end.

any_is_float([Var|Left], Info)->
  case hipe_icode:is_const(Var) of
    false ->
      case t_is_float(lookup_type(Var, Info)) of
	true -> true;
	false -> any_is_float(Left, Info)
      end;
    true ->
      case hipe_icode:const_value(Var) of
	X when is_float(X) -> true;
	_ -> any_is_float(Left, Info)
      end
  end;
any_is_float([], _Info) ->
  false.

remove_duplicates(List)->
  remove_duplicates(List, []).
remove_duplicates([X|Left], Acc)->
  case lists:member(X, Acc) of
    true ->
      remove_duplicates(Left, Acc);
    false ->
      remove_duplicates(Left, [X|Acc])
  end;
remove_duplicates([], Acc) ->
  Acc.

fun_to_fop(Fun)->
  case Fun of
    '+' -> fp_add;
    '-' -> fp_sub;
    '*' -> fp_mul;
    '/' -> fp_div;
    _ ->false
  end.


%% If there is a tagged version of this variable available we don't
%% have to tag the untagged version.

must_be_tagged(Var, Map)->
  case gb_trees:lookup(Var, Map) of
    none -> false;
    {value, {assigned, _}} -> false; 
    {value, Val} -> 
      case hipe_icode:is_fvar(Val) of
	true->
	  %% io:format("~w must be tagged since it is mapped to ~w\n", [Var, Val]),
	  true;
	false ->
	  false
      end
  end.


%% Converting to floating point variables

get_conv_instrs(Vars, Map, Info)->
  get_conv_instrs(Vars, Map, Info, []).

get_conv_instrs([Var|Left], Map, Info, Acc)->
  {_, Dst} = gb_trees:get(Var, Map),
  NewI = 
    case t_is_float(lookup_type(Var, Info)) of
      true ->
	case hipe_icode:is_const(Var) of
	  true ->
	    Tmp = hipe_icode:mk_new_var(),
	    [hipe_icode:mk_mov(Tmp, Var),
	     hipe_icode:mk_primop([Dst],unsafe_untag_float,[Tmp])];
	  false ->
	    [hipe_icode:mk_primop([Dst],unsafe_untag_float,[Var])]
	end;
      false ->
	[hipe_icode:mk_primop([Dst],conv_to_float,[Var])] 
    end,
  get_conv_instrs(Left, Map, Info, NewI++Acc);
get_conv_instrs([], _, _, Acc) ->
  Acc.


%% ------------------------------------------------------------ 
%% Defines and uses

defines(I)->
  hipe_icode:defines(I).

args(I)->
  hipe_icode:args(I).

uses(I)->
  keep_vars(hipe_icode:uses(I)).

keep_vars(Vars)->
  lists:filter(fun(X)->hipe_icode:is_var(X)end, Vars).


%% _________________________________________________________________
%%
%% Handling the state
%%

new_state(Cfg)->
  Start = hipe_icode_cfg:start(Cfg),  
  Info = case lists:keysearch(arg_type, 1, hipe_icode_cfg:info(Cfg)) of
	   false ->
	     Any = t_any(),
	     lists:foldl(fun(X, Tree)->gb_trees:insert(X, Any, Tree)end,
			 empty(), hipe_icode_cfg:params(Cfg));
	   {value,{_, ArgType}}->
	     add_arg_types(hipe_icode_cfg:params(Cfg), ArgType)
	 end,
  InfoMap = gb_trees:insert({Start, in}, Info, empty()),
  new_state(Cfg, InfoMap).

new_state(Cfg, InfoMap)->
  Start = hipe_icode_cfg:start(Cfg),  
  BlockMap = gb_trees:insert({inblock, Start}, false, empty()),
  #state{info_map=InfoMap, cfg=Cfg, block_map=BlockMap}.

add_arg_types(Args,Types)->
  add_arg_types(Args, Types, empty()).

add_arg_types([Arg|Args],[Type|Types], Acc)->
  add_arg_types(Args,Types, enter_type(Arg, Type, Acc));
add_arg_types([],[],Acc) ->
  Acc;
add_arg_types(A,B,_) ->
  exit({wrong_number_of_arguments, {A, B}}).

state__cfg(#state{cfg=Cfg})->
  Cfg.

state__succ(#state{cfg=Cfg}, Label)->
  hipe_icode_cfg:succ(hipe_icode_cfg:succ_map(Cfg), Label).

state__pred(#state{cfg=Cfg}, Label)->
  hipe_icode_cfg:pred(hipe_icode_cfg:pred_map(Cfg), Label).

state__redirect(S=#state{cfg=Cfg}, From, ToOld, ToNew)->
  NewCfg = hipe_icode_cfg:redirect(Cfg, From, ToOld, ToNew),
  S#state{cfg=NewCfg}.

state__bb(#state{cfg=Cfg}, Label)->
  hipe_icode_cfg:bb(Cfg, Label).
  
state__bb_update(S=#state{cfg=Cfg}, Label, BB)->
  NewCfg = hipe_icode_cfg:bb_update(Cfg, Label, BB),
  S#state{cfg=NewCfg}.

state__info_in(S, Label)->
  state__info(S, {Label, in}).

state__info_out(S, Label)->
  state__info(S, {Label, out}).

state__info(#state{info_map=IM}, Label)->
  case gb_trees:lookup(Label, IM) of
    {value, Info}-> Info;
    none -> empty()
  end.

state__info_in_update(S=#state{info_map=IM}, Label, Info)->
  case gb_trees:lookup({Label, in}, IM) of
    none ->
      S#state{info_map=gb_trees:enter({Label, in}, Info, IM)};
    {value, Info} ->
      fixpoint;
    {value, OldInfo} ->
      NewInfo = join_info_in(gb_trees:to_list(OldInfo), Info),
      S#state{info_map=gb_trees:enter({Label, in}, NewInfo, IM)}
  end.

state__info_out_update(S=#state{info_map=IM}, Label, Info)->
  S#state{info_map=gb_trees:enter({Label, out}, Info, IM)}.

state__map(S=#state{block_map=BM}, Label)->
  join_maps(state__pred(S, Label), BM).

state__map_update(S=#state{block_map=BM}, Label, Map, InBlock)->
  InBlockChanged = 
    case gb_trees:lookup({inblock_out, Label}, BM) of
      {value, InBlock} -> false;
      _OldInBlock -> 
	true
    end,
  MapChanged = 
    case gb_trees:lookup(Label, BM) of
      {value, Map} -> false;
      none -> true;
      {value, _OldMap} ->
	true
    end,
  case InBlockChanged orelse MapChanged of
    true ->
      NewBM1 = gb_trees:enter({inblock_out, Label}, InBlock, BM),
      NewBM2 = gb_trees:enter(Label, Map, NewBM1),
      S#state{block_map = NewBM2};
    false ->
      fixpoint
  end.

state__in_block(S=#state{block_map=BM}, Label)->
  Pred = state__pred(S, Label),
  InBlock = join_inblock(Pred, BM),
  NewBM = gb_trees:enter({inblock_in, Label}, InBlock, BM),
  {S#state{block_map=NewBM}, InBlock}.

state__get_in_block_in(#state{block_map=BM}, Label)->
  lookup({inblock_in, Label}, BM).

state__get_in_block_out(#state{block_map=BM}, Label)->
  lookup({inblock_out, Label}, BM).

print_state_info(State)->
  Labels = hipe_icode_cfg:reverse_postorder(state__cfg(State)),
  lists:foreach(
    fun(X)-> 
	io:format("Label ~w:\n", [X]),
	lists:foreach(
	  fun({Y, Z})->
	      io:format("\t~w: ~s\n", [Y, t_to_string(Z)])
	  end,
	  gb_trees:to_list(state__info_out(State, X)))
    end,
    Labels),
  io:format("=========================\n", []),
  ok.


new_worklist(#state{cfg=Cfg})->
  Start = hipe_icode_cfg:start(Cfg),
  {[Start], [], gb_sets:insert(Start, gb_sets:empty())}.

get_work({[Label|Left], List, Set})->
  {Label, {Left, List, gb_sets:delete(Label, Set)}};
get_work({[], [], _Set}) ->
  none;
get_work({[], List, Set}) ->
  get_work({lists:reverse(List), [], Set}).

add_work({List1, List2, Set}, [Label|Left])->
  case gb_sets:is_member(Label, Set) of
    true -> 
      add_work({List1, List2, Set}, Left);
    false -> 
      %%io:format("Added work: ~w\n", [Label]),
      NewSet = gb_sets:insert(Label, Set),
      add_work({List1, [Label|List2], NewSet}, Left)
  end;
add_work(WorkList, []) ->
  WorkList.
