%% -*- erlang-indent-level: 2 -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% ============================================================================
%%  Filename :  hipe_icode_ssa_const_prop.erl
%%  Authors  :  Daniel Luna, Erik Andersson
%%  Purpose  :  Perform sparse conditional constant propagation.
%%  Notes    :  Works on the control-flow graph.
%%
%%  History  : * 2003-03-05: Created.
%%             * 2003-08-11: Passed simple testsuite.
%%             * 2003-10-01: Passed compiler testsuite.
%% ============================================================================
%%
%% Exports: sparse_cond_const_propagate/1.
%%
%% ============================================================================
%%
%% TODO: 
%%
%% Take care of failures in call and replace operation with apropriate
%% failure.
%%
%% Handle ifs with non-binary operators
%%
%% We want multisets for easier (and faster) creation of env->ssa_edges
%% 
%% Maybe do things with restore_catch, pushcatch if possible
%%
%% Propagation of constant arguments when the some arguments are bottom
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_icode_ssa_const_prop).
-export([sparse_cond_const_propagate/1]).

-include("../main/hipe.hrl").

-define(CONST_PROP_MSG(Str,L),ok).
%%-define(CONST_PROP_MSG(Str,L),io:format(Str,L)).

-record(env, {cfg,
	      executable_flags,
	      lattice_values,
	      ssa_edges
	     }).
%% -define(DEBUG, 1).

%%-----------------------------------------------------------------------------

sparse_cond_const_propagate(CFG) ->
  Environment    = create_environment(CFG),
  StartEdge      = {hipe_icode_cfg:start(CFG), hipe_icode_cfg:start(CFG)},
  NewEnvironment = scc([StartEdge], [], Environment),
  NewCFG         = update_cfg(NewEnvironment),
  NewCFG.

%%-----------------------------------------------------------------------------
%%------------------- create_environment --------------------------------------

create_environment(CFG) ->
  #env{cfg              = CFG,
       executable_flags = gb_sets:empty(),
       lattice_values   = initialize_lattice(CFG),
       ssa_edges        = initialize_ssa_edges(CFG)
      }.

%%-----------------------------------------------------------------------------

initialize_lattice(CFG) ->
  Lattice    = gb_trees:empty(),
  Parameters = hipe_icode_cfg:params(CFG),
  Inserter   = fun(Parameter, Tree) ->
		   gb_trees:insert(Parameter, bottom, Tree) end,
  lists:foldl(Inserter, Lattice, Parameters).

%%-----------------------------------------------------------------------------

initialize_ssa_edges(CFG) ->
  IterateNodes =
    fun(Node, Tree1) ->
	IterateInstructions =
	  fun(Instruction, Tree2) ->
	      IterateArguments =
		fun(Argument, Tree3) ->
		    Data    = gb_trees:lookup(Argument, Tree3),
		    NewEdge = {Node, Instruction},
		    case Data of
		      none ->
			%% insert assumes key is not present
			gb_trees:insert(Argument,
					[NewEdge],
					Tree3);
		      {value, EdgeList} ->
			%% update assumes key is present
			gb_trees:update(Argument, 
					[NewEdge | EdgeList],
					Tree3)
		    end
		end,
	      Arguments = hipe_icode:uses(Instruction),
	      lists:foldl(IterateArguments, Tree2, Arguments)
	  end,
	Instructions = hipe_bb:code(hipe_icode_cfg:bb(CFG, Node)),
	lists:foldl(IterateInstructions, Tree1, Instructions)
    end,
  NodeList = hipe_icode_cfg:labels(CFG),
  lists:foldl(IterateNodes, gb_trees:empty(), NodeList).

%%-----------------------------------------------------------------------------
%%----------------------------------- scc -------------------------------------

scc([], [], Environment) ->
  Environment;

%% Take an element from the FlowWorkList and process it
scc([{Source,Destination} | FlowWorkList], SSAWorkList, Environment) ->
  case executable({Source, Destination}, Environment) of
    true ->
      scc(FlowWorkList, SSAWorkList, Environment);
    false ->
      Environment1          = mark_as_executable({Source, Destination},
						 Environment),
      Code                  = extract_code(Destination, Environment),
      {Environment2, Code1, ExtraSSA} =
	visit_phi_nodes(Code, Destination, Environment1, []),
      Predecessors          = predecessors(Destination, Environment2),
      Executable            = fun(Pred) ->
				  executable({Pred, Destination},
					     Environment2) end,
      ExecutablePreds       = length(lists:filter(Executable, Predecessors)),
      case (ExecutablePreds =< 1) of
	false ->
	  scc(FlowWorkList, ExtraSSA ++ SSAWorkList, Environment2);
	true ->
	  {MoreFlowDestinations, MoreSSAWork, Environment3} = 
	    visit_expressions(Code1, Environment2),
	  Zipper        = fun(Node) -> {Destination, Node} end,
	  MoreFlowWork  = lists:map(Zipper, MoreFlowDestinations),
	  FlowWorkList1 = MoreFlowWork ++ FlowWorkList,
	  SSAWorkList1  = ExtraSSA ++ MoreSSAWork ++ SSAWorkList,
	  scc(FlowWorkList1, SSAWorkList1, Environment3)
      end
  end;

%% Take an element from the SSAWorkList and process it
scc([], [{Node, Instruction} | SSAWorkList], Environment) ->
  case reachable(Node, Environment) of
    true ->
      case hipe_icode:is_phi(Instruction) of
	true ->
	  {Environment1, MoreSSA} = visit_phi(Instruction, Node, Environment),
	  scc([], MoreSSA ++ SSAWorkList, Environment1);
	false ->
	  {MoreFlowDestinations, MoreSSAWork, Environment1} = 
	    visit_expression(Instruction, Environment),
	  SSAWorkList1  = MoreSSAWork ++ SSAWorkList,
	  Zipper        = fun(Destination) -> {Node, Destination} end,
	  MoreFlowWork  = lists:map(Zipper, MoreFlowDestinations),
	  scc(MoreFlowWork, SSAWorkList1, Environment1)
      end;
    false ->
      scc([], SSAWorkList, Environment)
  end.

%%-----------------------------------------------------------------------------

visit_expressions(Instructions, Environment) ->
  visit_expressions(Instructions, Environment, [], []).

visit_expressions([], Environment, FlowWork, SSAWork) ->
  {FlowWork, SSAWork, Environment};
visit_expressions([Inst | Insts], Environment, FlowWork, SSAWork) ->
  {MoreFlowWork, MoreSSAWork, Environment1}
    = visit_expression(Inst, Environment),
  FlowWork1 = MoreFlowWork ++ FlowWork,
  SSAWork1  = MoreSSAWork  ++ SSAWork,
  visit_expressions(Insts, Environment1, FlowWork1, SSAWork1).

%%-----------------------------------------------------------------------------

visit_expression(Instruction, Environment) ->
  Arguments          = hipe_icode:args(Instruction),
  Lookup             = fun(Argument) ->
			   lookup_lattice_value(Argument, Environment)
		       end,
  EvaluatedArguments = lists:map(Lookup, Arguments),
  case hipe_icode:type(Instruction) of
    mov  ->
      visit_mov_or_fmov       (Instruction, EvaluatedArguments, Environment);
    fmov  ->
      visit_mov_or_fmov       (Instruction, EvaluatedArguments, Environment);
    'if' ->
      visit_if                (Instruction, EvaluatedArguments, Environment);
    goto ->
      visit_goto              (Instruction, EvaluatedArguments, Environment);
    type ->
      visit_type              (Instruction, EvaluatedArguments, Environment);
    call ->
      visit_call              (Instruction, EvaluatedArguments, Environment);
    switch_val ->
      visit_switch_val        (Instruction, EvaluatedArguments, Environment);
    switch_tuple_arity ->
      visit_switch_tuple_arity(Instruction, EvaluatedArguments, Environment);
    restore_catch ->
      visit_restore_catch     (Instruction, EvaluatedArguments, Environment);
    pushcatch ->
      visit_pushcatch         (Instruction, EvaluatedArguments, Environment);
    _ ->
      %% label, remove_catch, comment, return, fail
      {[], [], Environment}
  end.

%%-----------------------------------------------------------------------------

visit_pushcatch(Instruction, [], Environment) ->
  Label     = hipe_icode:pushcatch_label(Instruction), 
  Successor = hipe_icode:pushcatch_successor(Instruction),
  {[Label, Successor], [], Environment}.

%%-----------------------------------------------------------------------------

visit_restore_catch(Instruction, _Arguments, Environment) ->
  Destination1 = hipe_icode:restore_catch_type_dst(Instruction),
  {Environment1, SSAWork1} = update_lattice_value({Destination1, bottom},
						  Environment),
  Destination2 = hipe_icode:restore_catch_reason_dst(Instruction),
  {Environment2, SSAWork2} = update_lattice_value({Destination2, bottom},
						  Environment1),
  MoreSSAWork  = SSAWork1 ++ SSAWork2,
  {[], MoreSSAWork, Environment2}.

%%-----------------------------------------------------------------------------

visit_switch_val(Instruction, [Argument], Environment) ->
  Cases     = hipe_icode:switch_val_cases(Instruction),
  FailLabel = hipe_icode:switch_val_fail_label(Instruction),
  case Argument of
    bottom ->
      Second    = fun({_A,B}) -> B end,
      FlowWork  = lists:map(Second, Cases),
      FlowWork1 = [FailLabel | FlowWork],
      {FlowWork1, [], Environment};
    _ ->
      Target = get_switch_target(Cases, Argument, FailLabel),
      {[Target], [], Environment}
  end.

%%-----------------------------------------------------------------------------

visit_switch_tuple_arity(Instruction, [Argument], Environment) ->
  Cases     = hipe_icode:switch_tuple_arity_cases(Instruction),
  FailLabel = hipe_icode:switch_tuple_arity_fail_label(Instruction),
  case Argument of
    bottom ->
      Second    = fun({_A,B}) -> B end,
      FlowWork  = lists:map(Second, Cases),
      FlowWork1 = [FailLabel | FlowWork],
      {FlowWork1, [], Environment};
    Constant ->
      UnTagged = hipe_icode:const_value(Constant),
      case is_tuple(UnTagged) of
	true ->
	  Target = get_switch_target(Cases, size(UnTagged), FailLabel),
	  {[Target], [], Environment};
	false ->
	  {[FailLabel], [], Environment}
      end
  end.

%%-----------------------------------------------------------------------------

get_switch_target([], _Argument, FailLabel) ->
  FailLabel;

get_switch_target([{CaseValue, Target} | CaseList], Argument, FailLabel) ->
  case CaseValue == Argument of
    true ->
      Target;
    false ->
      get_switch_target(CaseList, Argument, FailLabel)
  end.

%%-----------------------------------------------------------------------------

visit_mov_or_fmov(Instruction, [SourceValue], Environment) ->
  Destination = case hipe_icode:type(Instruction) of
		  mov  -> hipe_icode:mov_dst(Instruction);
		  fmov -> hipe_icode:fmov_dst(Instruction)
		end,
  {Environment1, SSAWork} = update_lattice_value({Destination, SourceValue},
						 Environment),
  {[], SSAWork, Environment1}.

%%-----------------------------------------------------------------------------

visit_if(Instruction, Arguments, Environment) ->
  case evaluate_if(hipe_icode:if_op(Instruction), Arguments) of
    true ->
      TrueLabel  = hipe_icode:if_true_label(Instruction),
      FlowWork   = [TrueLabel];
    false ->
      FalseLabel = hipe_icode:if_false_label(Instruction),
      FlowWork   = [FalseLabel];
    bottom ->
      TrueLabel  = hipe_icode:if_true_label(Instruction),
      FalseLabel = hipe_icode:if_false_label(Instruction),
      FlowWork   = [TrueLabel, FalseLabel]
  end,
  {FlowWork, [], Environment}.

%%-----------------------------------------------------------------------------

visit_goto(Instruction, _Arguments, Environment) ->
  GotoLabel = hipe_icode:goto_label(Instruction),
  FlowWork  = [GotoLabel],
  {FlowWork, [], Environment}.

%%-----------------------------------------------------------------------------

visit_type(Instruction, [Value], Environment) ->
  case evaluate_type(hipe_icode:type_type(Instruction), Value) of
    true ->
      TrueLabel  = hipe_icode:type_true_label(Instruction),
      FlowWork   = [TrueLabel];
    false ->
      FalseLabel = hipe_icode:type_false_label(Instruction),
      FlowWork   = [FalseLabel];
    bottom ->
      TrueLabel  = hipe_icode:type_true_label(Instruction),
      FalseLabel = hipe_icode:type_false_label(Instruction),
      FlowWork   = [TrueLabel, FalseLabel]
  end,
  {FlowWork, [], Environment}.

%%-----------------------------------------------------------------------------

visit_call(Ins, Args, Environment) ->
  Dst = hipe_icode:call_dst(Ins),
  Fun = hipe_icode:call_fun(Ins),
  Fail = call_fail(Ins),
  Cont = call_continuation(Ins),
  visit_call(Dst, Args, Fun, Cont, Fail, Environment).

visit_call(Dst, Args, Fun, Cont, Fail, Environment) ->
  case lists:any(fun(X) -> (X == bottom) end, Args) of
    true ->
      FlowWork = Fail ++ Cont,
      {Environment1, SSAWork} =
	update_lattice_value({Dst, bottom}, Environment);
    false ->
      GetConst = fun(Argument) -> hipe_icode:const_value(Argument) end,
      ConstArgs   = lists:map(GetConst, Args),
      case catch evaluate_call_or_enter(ConstArgs, Fun) of
	{'EXIT', _} ->
	  FlowWork = Fail,
	  {Environment1, SSAWork} =
	    update_lattice_value({Dst, bottom}, Environment);
	bottom ->
	  FlowWork = Fail ++ Cont,
	  {Environment1, SSAWork} = 
	    update_lattice_value({Dst, bottom}, Environment);
	Constant ->
	  FlowWork = Cont,
	  {Environment1, SSAWork} =
	    update_lattice_value({Dst, Constant}, Environment)
      end
  end,
  {FlowWork, SSAWork, Environment1}.

%%-----------------------------------------------------------------------------

call_fail(I)->
  case hipe_icode:call_fail(I) of
    [] -> [];
    Label -> [Label]
  end.

call_continuation(I)->
  case hipe_icode:call_continuation(I) of
    [] -> [];
    Label -> [Label]
  end.
  
%%-----------------------------------------------------------------------------

%% Unary calls
evaluate_call_or_enter([Argument], Fun) ->
  case Fun of
    {erlang, hd, 1} ->
      hipe_icode:mk_const(hd(Argument));
    {erlang, tl, 1} ->
      hipe_icode:mk_const(tl(Argument));
    {unsafe_element, N} ->
      hipe_icode:mk_const(element(N, Argument));
    {erlang, atom_to_list, 1} ->
      hipe_icode:mk_const(atom_to_list(Argument));
    {erlang, list_to_atom, 1} ->
      hipe_icode:mk_const(list_to_atom(Argument));
    {erlang, tuple_to_list, 1} ->
      hipe_icode:mk_const(tuple_to_list(Argument));
    {erlang, list_to_tuple, 1} ->
      hipe_icode:mk_const(list_to_tuple(Argument));
    {erlang, length, 1} ->
      hipe_icode:mk_const(length(Argument));
    {erlang, size, 1} ->
      hipe_icode:mk_const(size(Argument));
    {erlang, abs, 1} ->
      hipe_icode:mk_const(abs(Argument));
    {erlang, round, 1} ->
      hipe_icode:mk_const(round(Argument));
    {erlang, trunc, 1} ->
      hipe_icode:mk_const(trunc(Argument));
    mktuple ->
      hipe_icode:mk_const(list_to_tuple([Argument]));
    unsafe_untag_float ->
      hipe_icode:mk_const(float(Argument));
    conv_to_float ->
      hipe_icode:mk_const(float(Argument));
    fnegate ->
      hipe_icode:mk_const(0.0 - Argument);
    'bnot' ->
      hipe_icode:mk_const(Argument);
    _ ->
      bottom
  end;

%% Binary calls
evaluate_call_or_enter([Argument1,Argument2], Fun) ->
  case Fun of
    '+' ->
      hipe_icode:mk_const(Argument1 + Argument2);
    '-' ->
      hipe_icode:mk_const(Argument1 - Argument2);
    '*' ->
      hipe_icode:mk_const(Argument1 * Argument2);
    '/' ->
      hipe_icode:mk_const(Argument1 / Argument2);
    {erlang, 'div', 2} ->
      hipe_icode:mk_const(Argument1 div Argument2);
    {erlang, 'rem', 2} ->
      hipe_icode:mk_const(Argument1 rem Argument2);
    'bsl' ->
      hipe_icode:mk_const(Argument1 bsl Argument2);
    'bsr' ->
      hipe_icode:mk_const(Argument1 bsl Argument2);
    'band' ->
      hipe_icode:mk_const(Argument1 band Argument2);
    'bor' ->
      hipe_icode:mk_const(Argument1 bor Argument2);
    'bxor' ->
      hipe_icode:mk_const(Argument1 bxor Argument2);
    fp_add ->
      hipe_icode:mk_const(Argument1 + Argument2);
    fp_sub ->
      hipe_icode:mk_const(Argument1 - Argument2);
    fp_mul ->
      hipe_icode:mk_const(Argument1 * Argument2);
    fp_div ->
      hipe_icode:mk_const(Argument1 / Argument2);
    cons ->
      hipe_icode:mk_const([Argument1 | Argument2]);
    {erlang, '++', 2} ->
      hipe_icode:mk_const(Argument1 ++ Argument2);
    {erlang, '--', 2} ->
      hipe_icode:mk_const(Argument1 -- Argument2);
    {erlang, append_element, 2} ->
      hipe_icode:mk_const(erlang:append_element(Argument1, Argument2));
    mktuple -> 
      hipe_icode:mk_const(list_to_tuple([Argument1,Argument2]));
    _Other ->
      %% io:format("In ~w(~w,~w)~n", [_Other,Argument1,Argument2]),
      bottom
  end;

%% The rest of the calls
evaluate_call_or_enter(Arguments, Fun) ->
  case Fun of
    mktuple ->
      hipe_icode:mk_const(list_to_tuple(Arguments));
    {erlang, setelement, 3} ->
      [Argument1, Argument2, Argument3] = Arguments,
      hipe_icode:mk_const(setelement(Argument1, Argument2, Argument3));
    _ ->
      bottom
  end.

%%-----------------------------------------------------------------------------

evaluate_if(Conditional, [Argument1, Argument2]) -> 
  case ((Argument1 == bottom) or (Argument2 == bottom)) of
    true  -> bottom;
    false -> evaluate_if_const(Conditional, Argument1, Argument2)
  end;
evaluate_if(_Conditional, _Arguments) ->
  bottom.

%%-----------------------------------------------------------------------------

evaluate_if_const(Conditional, Argument1, Argument2) ->
  case Conditional of
    '=:=' -> Argument1 =:= Argument2;
    '=='  -> Argument1 ==  Argument2;
    '=/=' -> Argument1 =/= Argument2;
    '/='  -> Argument1  /= Argument2;
    '<'   -> Argument1  <  Argument2;
    '>='  -> Argument1  >= Argument2;
    '=<'  -> Argument1 =<  Argument2;
    '>'   -> Argument1  >  Argument2;
    _     -> bottom
  end.

%%-----------------------------------------------------------------------------

evaluate_type(Type, Val) ->
  case Val of
    bottom -> bottom;
    _      -> evaluate_type_const(Type, hipe_icode:const_value(Val))
  end.

%%-----------------------------------------------------------------------------

evaluate_type_const(Type, Arg) ->
  case {Type, Arg} of
    {nil,    []   }  -> true;
    {nil,    _    }  -> false;
    {cons,   [_|_]}  -> true;
    {cons,   _    }  -> false;
    {{tuple, N}, T} when is_tuple(T), size(T) == N -> true;
    {atom,       A} when is_atom(A) -> true;
    {{atom, A},       A} when is_atom(A) -> true;
    _                -> bottom
  end.

%%-----------------------------------------------------------------------------

visit_phi_nodes([], CurrentNode, _Environment, _SSAWork) ->
  ?EXIT({"visit_phi_nodes: Basic block contains no code", CurrentNode});
visit_phi_nodes([Inst | Insts], Node, Environment, SSAWork) ->
  case hipe_icode:is_phi(Inst) of
    true ->
      {Environment1, NewSSA} = visit_phi(Inst, Node, Environment),
      visit_phi_nodes(Insts, Node, Environment1, NewSSA ++ SSAWork);
    false ->
      {Environment, [Inst | Insts], SSAWork}
  end.

%%-----------------------------------------------------------------------------

visit_phi(PhiInstruction, CurrentNode, Environment) ->
  ArgumentList = hipe_icode:phi_arglist(PhiInstruction),
  Value        = get_phi_value(ArgumentList, CurrentNode, Environment, top),
  Name         = hipe_icode:phi_dst(PhiInstruction),
  {Environment1, SSAWork} = update_lattice_value({Name, Value}, Environment),
  {Environment1, SSAWork}.

%%-----------------------------------------------------------------------------

get_phi_value([], _CurrentNode, _Environment, CurrentValue) ->
  CurrentValue;
get_phi_value([{PredecessorNode, Variable}| ArgumentList],
	      CurrentNode,
	      Environment,
	      CurrentValue) ->
  case executable({PredecessorNode, CurrentNode}, Environment) of
    true ->
      NewValue = lookup_lattice_value(Variable, Environment),
      case NewValue of
	bottom ->
	  bottom;
	top ->
	  get_phi_value(ArgumentList, CurrentNode, Environment, CurrentValue);
	_ ->
	  case CurrentValue of
	    top ->
	      get_phi_value(ArgumentList, CurrentNode, Environment, NewValue);
	    _ ->
	      case (NewValue == CurrentValue) of
		true ->
		  get_phi_value(ArgumentList,
				CurrentNode,
				Environment,
				NewValue);
		false ->
		  bottom
	      end
	  end
      end;
    false ->
      get_phi_value(ArgumentList, CurrentNode, Environment, CurrentValue)
  end.

%%-----------------------------------------------------------------------------
%%------------------------------- update_cfg ----------------------------------

update_cfg(Environment) ->
  NodeList = get_nodelist(Environment),
  CFG1     = update_nodes(NodeList, Environment),
  CFG2     = hipe_icode_cfg:remove_unreachable_code(CFG1),
  CFG2.

%%-----------------------------------------------------------------------------

update_nodes([], Environment) ->
  get_cfg(Environment);
update_nodes([Node | NodeList], Environment) ->
  case reachable(Node, Environment) of
    true ->
      Instructions    = extract_code(Node, Environment),
      Updater         = fun(Instruction) ->
			    update_instruction(Instruction, Environment) end,
      NewInstructions = lists:flatmap(Updater, Instructions),
      Environment1    = update_code(Node, NewInstructions, Environment);
    false -> 
      Environment1    = Environment
  end,      
  update_nodes(NodeList, Environment1).

%%-----------------------------------------------------------------------------

update_instruction(Instruction, Environment) ->
  case hipe_icode:type(Instruction) of
    call ->
      update_call(Instruction, Environment);
    enter ->
      update_enter(Instruction, Environment);
    'if' ->
      update_if(Instruction, Environment);
    mov ->
      update_mov(Instruction, Environment);
    phi ->
     update_phi(Instruction, Environment);
    switch_val ->
      update_switch_val(Instruction, Environment);
    type ->
      update_type(Instruction, Environment);
    switch_tuple_arity ->
      update_switch_tuple_arity(Instruction, Environment);
    _ ->
      %% goto, comment, label, return, restore_catch, remove_catch,
      %% pushcatch, fail
      %% We could but don't handle: catch?, fail?
      [Instruction]
  end.

%%-----------------------------------------------------------------------------

update_call(Instruction, Environment) ->
  DestinationList = hipe_icode:call_dst(Instruction),
  case DestinationList of
    [Destination] ->
      case lookup_lattice_value(Destination, Environment) of
	bottom ->
	  [Instruction];
	X -> 
	  NewInstructions =
	    case is_call_to_fp_op(Instruction) of
	      true ->
		TmpIns = 
		  hipe_icode:call_fun_update(Instruction, unsafe_untag_float),
		[hipe_icode:call_args_update(TmpIns, [X])];
	      false ->
		case hipe_icode:call_continuation(Instruction) of
		  [] ->
		    [hipe_icode:mk_mov(Destination, X)];
		  ContinuationLabel ->
		    [hipe_icode:mk_mov(Destination, X),
		     hipe_icode:mk_goto(ContinuationLabel)]
		end
	    end,
	  ?CONST_PROP_MSG("call: ~w ---> ~w\n",
			  [Instruction, NewInstructions]),
	  NewInstructions
      end;
    %% [] ->  %% No destination; we don't touch this
    %% List-> %% Means register allocation; not implemented at this point
    _ ->
      [Instruction]
  end.

%%-----------------------------------------------------------------------------

is_call_to_fp_op(Ins)->
  case hipe_icode:call_fun(Ins) of
    fp_add             -> true;
    fp_sub             -> true;
    fp_mul             -> true;
    fp_div             -> true;
    fnegate            -> true;
    conv_to_float      -> true;
    unsafe_untag_float -> true;
    _                  -> false
  end.

%%-----------------------------------------------------------------------------

update_enter(Ins, Env) ->
  Args = hipe_icode:enter_args(Ins),
  EvalArgs = [lookup_lattice_value(X, Env) || X<-Args],
  Fun = hipe_icode:enter_fun(Ins),
  case lists:any(fun(X) -> (X == bottom) end, EvalArgs) of
    true ->
      [Ins];
    false ->
      ConstVals = [hipe_icode:const_value(X)||X<-EvalArgs],
      case catch evaluate_call_or_enter(ConstVals, Fun) of
	{'EXIT', _} -> 
	  [Ins];
	bottom -> 
	  [Ins];
	Const ->
	  Dst = hipe_icode:mk_new_var(),
	  [hipe_icode:mk_mov(Dst, Const),
	   hipe_icode:mk_return([Dst])]
      end
  end.
  
%%-----------------------------------------------------------------------------

update_if(Instruction, Environment) ->
  Arguments          = hipe_icode:if_args(Instruction),
  Lookup             = fun(Argument) ->
			   lookup_lattice_value(Argument, Environment)
		       end,
  EvaluatedArguments = lists:map(Lookup, Arguments),
  case evaluate_if(hipe_icode:if_op(Instruction), EvaluatedArguments) of
    true ->
      TrueLabel  = hipe_icode:if_true_label(Instruction),
      ?CONST_PROP_MSG("ifT: ~w ---> goto ~w\n", [Instruction, TrueLabel]),
      [hipe_icode:mk_goto(TrueLabel)];
    false ->
      FalseLabel = hipe_icode:if_false_label(Instruction),
      ?CONST_PROP_MSG("ifF: ~w ---> goto ~w\n", [Instruction, FalseLabel]),
      [hipe_icode:mk_goto(FalseLabel)];
    bottom ->
      [Instruction]
  end.

%%-----------------------------------------------------------------------------

update_mov(Instruction, Environment) ->
  Destination = hipe_icode:mov_dst(Instruction),
  case lookup_lattice_value(Destination, Environment) of
    bottom ->
      [Instruction];
    X ->
      case hipe_icode:mov_src(Instruction) of
	X ->
	  [Instruction];
	_ ->
	  ?CONST_PROP_MSG("mov: ~w ---> ~w\n", [Instruction, X]),
	  [hipe_icode:mov_src_update(Instruction, X)]
      end
      %% == [hipe_icode:mk_mov(Destination, X)]
  end.

%%-----------------------------------------------------------------------------

update_phi(Instruction, Environment) ->
  Destination = hipe_icode:phi_dst(Instruction),
  case lookup_lattice_value(Destination, Environment) of
    bottom ->
      [Instruction];
    X ->
      ?CONST_PROP_MSG("phi: ~w ---> ~w\n", [Instruction, X]),
      [hipe_icode:mk_mov(Destination, X)]
  end.

%%-----------------------------------------------------------------------------

update_type(Instruction, Environment) ->
  Argument          = hipe_icode:type_var(Instruction),
  EvaluatedArgument = lookup_lattice_value(Argument, Environment),
  case evaluate_type(hipe_icode:type_type(Instruction), EvaluatedArgument) of
    true ->
      TrueLabel  = hipe_icode:type_true_label(Instruction),
      ?CONST_PROP_MSG("typeT: ~w ---> goto ~w\n", [Instruction, TrueLabel]),
      [hipe_icode:mk_goto(TrueLabel)];
    false ->
      FalseLabel = hipe_icode:type_false_label(Instruction),
      ?CONST_PROP_MSG("typeF: ~w ---> goto ~w\n", [Instruction, FalseLabel]),
      [hipe_icode:mk_goto(FalseLabel)];
    bottom ->
      [Instruction]
  end.

%%-----------------------------------------------------------------------------

update_switch_val(Instruction, Environment) ->
  Argument = hipe_icode:switch_val_arg(Instruction),
  Value    = lookup_lattice_value(Argument, Environment),
  case Value of
    bottom ->
      [Instruction];
    _ ->
      Cases     = hipe_icode:switch_val_cases(Instruction),
      FailLabel = hipe_icode:switch_val_fail_label(Instruction),
      Target    = get_switch_target(Cases, Value, FailLabel),
      ?CONST_PROP_MSG("sv: ~w ---> goto ~w\n", [Instruction, Target]),
      [hipe_icode:mk_goto(Target)]
  end.

%%-----------------------------------------------------------------------------

update_switch_tuple_arity(Instruction, Environment) ->
  Argument = hipe_icode:switch_tuple_arity_arg(Instruction),
  Value    = lookup_lattice_value(Argument, Environment),
  case Value of
    bottom ->
      [Instruction];
    Constant ->
      UnTagged = hipe_icode:const_value(Constant),
      case is_tuple(UnTagged) of
	true ->
	  Cases     = hipe_icode:switch_tuple_arity_cases(Instruction),
	  FailLabel = hipe_icode:switch_tuple_arity_fail_label(Instruction),
	  Target = get_switch_target(Cases, size(UnTagged), FailLabel),
	  ?CONST_PROP_MSG("sta: ~w ---> goto ~w\n", [Instruction, Target]),
	  [hipe_icode:mk_goto(Target)];
	false ->
	  [Instruction]
          %% TODO: Can the above be replaced with below??? Perhaps
	  %% together with some sort of "generate failure".
	  %% [hipe_icode:mk_goto(FailLabel)]
      end
  end.

%%-----------------------------------------------------------------------------

update_code(Node, NewInstructions, Environment) ->
  CFG                 = Environment#env.cfg,
  BB                  = hipe_icode_cfg:bb(CFG, Node),
  OrderedInstructions = put_phi_nodes_first(NewInstructions),
  NewBB               = hipe_bb:code_update(BB, OrderedInstructions),
  NewCFG              = hipe_icode_cfg:bb_update(CFG, Node, NewBB),
  Environment#env{cfg = NewCFG}.

%%-----------------------------------------------------------------------------

put_phi_nodes_first(Instructions) ->
  {PhiInstructions, OtherInstructions} = 
    partition(fun(X) -> hipe_icode:is_phi(X) end, Instructions),
  PhiInstructions ++ OtherInstructions.

%%-----------------------------------------------------------------------------

partition(Function, List) ->
  partition(Function, List, [], []).

partition(_Function, [], True, False) ->
  {lists:reverse(True), lists:reverse(False)};

partition(Function, [Hd | Tail], True, False) ->
  case Function(Hd) of
    true ->
      partition(Function, Tail, [Hd | True], False);
    false ->
      partition(Function, Tail, True, [Hd | False])
  end.

%%-----------------------------------------------------------------------------
%%------------------------------ environment ----------------------------------

reachable(Node, Environment) ->
  Predecessors = predecessors(Node, Environment),
  Executable   = fun(Pred) -> executable({Pred, Node}, Environment) end,
  lists:any(Executable, Predecessors).

%%-----------------------------------------------------------------------------

mark_as_executable(Edge, Environment) ->
  ExecutableFlags  = Environment#env.executable_flags,
  ExecutableFlags1 = gb_sets:add(Edge, ExecutableFlags),
  Environment#env{executable_flags = ExecutableFlags1}.

%%-----------------------------------------------------------------------------

extract_code(Node, Environment) ->
  CFG = Environment#env.cfg,
  case hipe_icode_cfg:bb(CFG, Node) of
    not_found -> io:format("Could not find label ~w\n", [Node]),
		 [];
    BB -> hipe_bb:code(BB)
  end.

%%-----------------------------------------------------------------------------

predecessors(Node, Environment) ->
  CFG = Environment#env.cfg,
  hipe_icode_cfg:pred(hipe_icode_cfg:pred_map(CFG), Node).

%%-----------------------------------------------------------------------------

executable(Edge, Environment) ->
  ExecutableFlags = Environment#env.executable_flags,
  gb_sets:is_member(Edge, ExecutableFlags).

%%-----------------------------------------------------------------------------

lookup_lattice_value(X, Environment) ->
  LatticeValues = Environment#env.lattice_values,
  case hipe_icode:is_const(X) of
    true ->
      X;
    false ->
      case gb_trees:lookup(X, LatticeValues) of
	none ->
	  ?WARNING_MSG("Earlier compiler steps generated erroneous " 
		       "code for X = ~w. We are taking care of this.\n",[X]),
	  bottom;
	{value, top} ->
	  ?EXIT("lookup_lattice_value, top");
	{value, Y} ->
	  Y
      end
  end.

%%-----------------------------------------------------------------------------

update_lattice_value({[], _NewValue}, Environment) ->
  {Environment, []};
update_lattice_value({Names, NewValue}, Environment) when is_list(Names) ->
  Update = 
    fun(Dst, {Env, SSA}) ->
	{NewEnv, NewSSA} = 
	  update_lattice_value({Dst, NewValue}, Env),
	{NewEnv, SSA++NewSSA}
    end,
  lists:foldl(Update, {Environment, []}, Names);
update_lattice_value({Name, NewValue}, Environment) ->
  LatticeValues  = Environment#env.lattice_values,
  case gb_trees:lookup(Name, LatticeValues) of
    none ->
      LatticeValues1 = gb_trees:insert(Name, NewValue, LatticeValues),
      SSAWork = lookup_ssa_edges(Name, Environment);
    {value, NewValue} ->
      LatticeValues1 = LatticeValues,
      SSAWork = [];
    {value, _} ->
      LatticeValues1 = gb_trees:update(Name, NewValue, LatticeValues),
      SSAWork = lookup_ssa_edges(Name, Environment)
  end,
  {Environment#env{lattice_values = LatticeValues1}, SSAWork}.

%%-----------------------------------------------------------------------------

lookup_ssa_edges(Variable, Environment) ->
  SSAEdges = Environment#env.ssa_edges,
  case gb_trees:lookup(Variable, SSAEdges) of
    {value, X} ->
      X;
    _ -> % Unused variable
      []
  end.

%%-----------------------------------------------------------------------------

get_nodelist(Environment) ->
  CFG = Environment#env.cfg,
  hipe_icode_cfg:labels(CFG).

%%-----------------------------------------------------------------------------

get_cfg(Environment) ->
  Environment#env.cfg.

%%-----------------------------------------------------------------------------

-ifdef(DEBUG).

%%-----------------------------------------------------------------------------
%%---------------------------------- DEBUG ------------------------------------

error(Text) ->
  error(Text, []).

error(Text, Data) ->
  io:format("Internal compiler error in ~w\n",[?MODULE]),
  io:format(Text, Data),
  io:format("\n\n"),
  halt().

%%-----------------------------------------------------------------------------

print_environment(Environment) ->
  io:format("============================================================\n"),
  io:format("Executable flags: "),
  print_executable_flags(Environment#env.executable_flags),
  io:format("Lattice values --->\n"),
  print_lattice_values(Environment#env.lattice_values),
  io:format("SSA edges --->\n"),
  print_ssa_edges(Environment#env.ssa_edges),
  io:format("============================================================\n").
  
%%-----------------------------------------------------------------------------

print_executable_flags(ExecutableFlags) ->
  ListOfFlags = gb_sets:to_list(ExecutableFlags),
  Printer     = fun ({Source, Destination}) -> 
		    io:format("(~w, ~w), ", [Source, Destination]) end,
  lists:foreach(Printer, ListOfFlags),
  io:format("()\n").

%%-----------------------------------------------------------------------------

print_lattice_values(LatticeValues) ->
  ListOfLatticeValues = gb_trees:to_list(LatticeValues),
  Printer             = fun ({Key, Value}) ->
			    io:format("~w = ~w\n", [Key, Value]) end,
  lists:foreach(Printer, ListOfLatticeValues).

%%-----------------------------------------------------------------------------

print_ssa_edges(SSAEdges) ->
  ListOfSSAEdges = gb_trees:to_list(SSAEdges),
  Printer        = fun ({Key, Value}) ->
		       io:format("~w: ~w\n", [Key, Value]) end,
  lists:foreach(Printer, ListOfSSAEdges).

%%-----------------------------------------------------------------------------

-endif.	%% DEBUG

%%----------------------------- End of file -----------------------------------
