
%%%----------------------------------------------------------------------
%%% File    : hipe_x86_postpass.erl
%%% Author  : Christoffer Vikstrm <chvi3471@student.uu.se>
%%% Purpose : Contain postpass optimisations for x86-assembler code.
%%% Created : 5 Aug 2003 by Christoffer Vikstrm <chvi3471@student.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_x86_postpass).
-author('chvi3471@student.uu.se').

-include("hipe_x86.hrl").

-export([postpass/1]).

-define(DO_PEEP, true).      % Turn on/off peephole optimizations
-define(DO_LOGGING, false).  % Change this to get peephole opts logged to..
-define(LOG_FILE, "peepcount.txt"). % ..this file..
-define(LOG_DIR, "/home/chvi3471/x-jobb/otp/tests/"). % ..in this dir


%%>----------------------------------------------------------------------<
%  Procedure : postpass/1
%  Purpose   : Function that performs a nr of postpast optimizations on
%              the hipe x86-assembler code before it is encoded and loaded.
%  Arguments : Defun - Function definition. Contain all assembler code of 
%              a function in a list.
%  Return    : An optimized defun-record.
%  Notes     : 
%%>----------------------------------------------------------------------<
postpass(Defun) ->
    #defun{mfa=MFA, code=Code0} = Defun,
    printLogHeader(MFA),           
    Code1 = expand(Code0),      % Expand pseudo instructions
    Code2 = peep(Code1),        % Do peephole optimizations
    Code3 = peepN(Code2),       % Do necessary peephole optimizations
    Defun#defun{code=Code3}.


%%>----------------------------------------------------------------------<
%  Procedure : peep/1
%  Purpose   : Function that does peephole optimizations. It works by 
%              moving a window over the code and looking at a sequence of 
%              a few instructions. Replaces sequences of instructions
%              with single ones and removes unnecesary instructions.
%  Arguments : Insns   - List of pseudo x86-assembler records.
%  Return    : An optimized list of pseudo x86-assembler records with 
%              (hopefully) fewer or faster instructions.
%  Notes     : Creates a file in the users home directory that contain 
%              analysis information, if macro ?DO_LOGGING i set to true.
%%>----------------------------------------------------------------------< 
peep(Insns) -> 
    case ?DO_PEEP of
	true ->
	    peep(Insns, [], []);
	_ ->
	    Insns
    end.

%% MoveSelf related peep-opts 
%% ------------------------------
peep([#fmov{src=Src, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, Res, [moveSelf1|Lst]);
peep([I=#fmov{src=Src, dst=Dst}, 
      #fmov{src=Dst, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, [I|Res], [moveSelf2|Lst]);
peep([#movsx{src=Src, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, Res, [moveSelf3|Lst]);
peep([I=#movsx{src=Src, dst=Dst}, 
      #movsx{src=Dst, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, [I|Res], [moveSelf4|Lst]);
peep([#movzx{src=Src, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, Res, [moveSelf5|Lst]);
peep([I=#movzx{src=Src, dst=Dst}, 
      #movzx{src=Dst, dst=Src} | Insns], Res,Lst) ->
    peep(Insns, [I|Res], [moveSelf6|Lst]);
peep([#cmovcc{src=Src, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, Res, [moveSelf7|Lst]);
peep([I=#cmovcc{src=Src, dst=Dst}, 
      #cmovcc{src=Dst, dst=Src}|Insns], Res,Lst) -> 
    peep(Insns, [I|Res], [moveSelf8|Lst]);
peep([#move{src=Src, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, Res, [moveSelf9|Lst]);
peep([I=#move{src=Src, dst=Dst}, 
      #move{src=Dst, dst=Src} | Insns], Res,Lst) -> 
    peep(Insns, [I|Res], [moveSelf0|Lst]);

%% ElimBinALMDouble
%% ----------------
peep([Move, BinAL = #alu{aluop=OP, src=Src, dst=Dst}|Insns], Res, Lst) ->
    case Move of
	#move{src=S, dst=D} ->
	    case (D==Dst) and (S==Src) of
		true ->
		    BinALNew = #alu{aluop=OP, src=D, dst=Dst},
		    peep(Insns, [BinALNew|Res], [elimBinALMDouble|Lst]);
		false ->
		    peep(Insns, [BinAL,Move|Res], Lst)
	    end;
	#movsx{src=S, dst=D} ->
	    case (D==Dst) and (S==Src) of
		true ->
		    BinALNew = #alu{aluop=OP, src=D, dst=Dst},
		    peep(Insns, [BinALNew|Res], [elimBinALMDouble|Lst]);
		false ->
		    peep(Insns, [BinAL,Move|Res], Lst)
	    end;
	#movzx{src=S, dst=D} ->
	    case (D==Dst) and (S==Src) of
		true ->
		    BinALNew = #alu{aluop=OP, src=D, dst=Dst},
		    peep(Insns, [BinALNew|Res], [elimBinALMDouble|Lst]);
		false ->
		    peep(Insns, [BinAL,Move|Res], Lst)
	    end;
	_ ->
	    peep(Insns, [BinAL,Move|Res], Lst)
	end;

%% ElimFBinDouble
%% --------------
peep([M = #fmov{src=Src1, dst=Dst1}, 
      B = #fp_binop{op=Op, src=Src2, dst=Dst2}|Insns], Res, Lst) ->
    case (Src1==Src2) and (Dst1==Dst2) of
	true ->
	    peep(Insns, 
		 [#fp_binop{op=Op,src=Dst1,dst=Dst2}|Res], 
		 [elimFBinDouble|Lst]);
	false ->
	    peep(Insns, [B,M|Res], Lst)
    end;

%% CommuteBinALMD
%% --------------
peep([M = #move{src=Src1, dst=Dst1}, 
      B = #alu{aluop=Op,src=Src2,dst=Dst2}|Insns], Res, Lst) ->
    
    case (Src1==#x86_imm{}) and (Src2/=#x86_imm{}) and (Dst1==Dst2) of
	true ->
	    case (Op=='add') or (Op=='and') or (Op=='or') or (Op=='xor') of
		true ->
		    peep(Insns, 
			 [#alu{aluop=Op,src=Src1,dst=Dst2},
			  #move{src=Src2, dst=Dst1}|Res], 
			 [commuteBinALMD|Lst]);
		false ->
		    peep(Insns, [B,M|Res], Lst)
	    end;
	false ->
	    peep(Insns, [B,M|Res], Lst)
    end;

%% ElimMDPow2
%% ----------
%% Unnecessary since BEAM converts all multiplications to shifts.
% peep([B = #alu{aluop=Op,src=#x86_imm{value=Val},dst=Dst}|Insns], Res, Lst) ->
%     {IsLog2, Size, Sign} = log2(Val),
%     case ((Op == imul) or (Op == idiv)) and IsLog2 of
% 	true ->
% 	    Sh = case Sign of positive -> 'bsl'; negative -> 'bsr' end,
% 	    io:format("ElimMDPow2!!!!~n", []),
% 	    peep(Insns, 
% 		 [#shift{shiftop=Sh, src=#x86_imm{value=Size}, dst=Dst}|Res], 
% 		 [elimMDPow2|Lst]);
% 	false ->
% 	    peep(Insns, [B|Res], Lst)
%     end;

%% ElimCmp0
%% --------
%% Since the assembler instruction 'test' is not implemented, this is not
%% usable.   
% peep([C=#cmp{src=Src, dst=Dst},J=#jcc{cc=Cond, label=Lab}|Insns],Res,Lst) ->
%     case ((Src = #x86_imm{value=0}) or (Dst = #x86_imm{value=0})) and
% 	((Cond == 'eq') or (Cond == 'neq')) of
% 	true ->
% 	    Src2 = case Src of #x86_imm{value=0} -> Src; O -> Dst end, 
% 	    Cond2 = case Cond of 'eq' -> 'z'; 'neq' -> 'nz' end,
% 	    Test = #test{src=Src2, dst=#x86_imm{value=0}},
% 	    Jump = #jcc{cc=Cond2, label=Lab},
% 	    peep(Insns, [Jump, Test|Res], [elimCmp0|Lst]);
% 	False ->
% 	    peep(Insns, [J,C|Res], Lst)
%     end;



%% ElimCmpTest
%% -----------
%%
%% peep([C=#test{} | Insns],Res,Lst) -> % This case could also be tested
%% but it need instruction #test{} to be implemented and supported by the 
%% assembler. As it is now, #test{}-instruction is never created by the 
%% instruction-selector.
peep([C=#cmp{} | Insns],Res,Lst) ->
    case check(Insns) of
	#jcc{} ->
	    peep(Insns, [C|Res], Lst);
	#jmp_fun{} ->
	    peep(Insns, [C|Res], Lst);
	#jmp_label{} ->
	    peep(Insns, [C|Res], Lst);
	#jmp_switch{} ->
	    peep(Insns, [C|Res], Lst);
	#cmovcc{} ->
	    peep(Insns, [C|Res], Lst);
	#ret{} ->
	    peep(Insns, [C|Res], Lst);
	_ ->
	    peep(Insns, Res, [elimCmpTest|Lst])
    end;


%% Standard list recursion clause
%% ------------------------------
peep([I | Insns], Res, Lst) ->
     peep(Insns, [I|Res], Lst);

%% Base case. Optionally prints an optimization log
%% ------------------------------------------------
peep([], Res, Lst) ->
    case ?DO_LOGGING of
	true ->
	    printLst(Lst);
	_ ->
	    ok
    end,
    lists:reverse(Res). 

%% Simple goto elimination (vital, dont know why..?)
%% -------------------------------------------------
peepN(Insns) -> peepN(Insns, [], []).
peepN([#jmp_label{label=Label}, I = #label{label=Label}|Insns], Res,Lst) ->
     peepN([I|Insns], Res, [nearGotoElim|Lst]);
peepN([I | Insns], Res, Lst) ->
     peepN(Insns, [I|Res], Lst);
peepN([], Res, Lst) ->
    printLst(Lst),
    lists:reverse(Res). 



%%>----------------------------------------------------------------------<
%  Procedure : expand/1
%  Purpose   : Expands pseudo instructions.
%  Arguments : Insns - An x86-instruction list.
%  Return    : An optimized instruction list.
%  Notes     : 
%%>----------------------------------------------------------------------<
expand(Insns) -> expand(Insns, []).
expand([I|Tail], Res) ->
    case I of
	#pseudo_jcc{cc=Cc,true_label=TrueLab,false_label=FalseLab} ->
	    expand(Tail, [hipe_x86:mk_jmp_label(FalseLab),
			  hipe_x86:mk_jcc(Cc, TrueLab) | Res]);
	#pseudo_tailcall_prepare{} ->
	    expand(Tail, Res);
	_ ->
	    expand(Tail, [I|Res])
    end;
expand([], Res) -> lists:reverse(Res).


%%>----------------------------------------------------------------------<
%  Procedure : printLogHeader/1
%  Purpose   : Prints the headers for each function compiled.
%  Arguments : MFA -  #x86_mfa{} record
%  Return    : unit
%  Notes     : 
%%>----------------------------------------------------------------------<
printLogHeader(MFA) ->
    case ?DO_LOGGING of
	true ->
	    {x86_mfa, M, F, _} = MFA,
	    {ok, Dir} = file:get_cwd(),
	    file:set_cwd(?LOG_DIR),
	    {ok, File} = file:open(?LOG_FILE, [read, append]),
	    io:format(File, 
		      "\nModule: ~w  Function: ~w" ++ 
		      "\n>=========================<\n", 
		      [M, F]),
	    file:set_cwd(Dir),
	    file:close(File);
	_ ->
	    ok
    end.


%%>----------------------------------------------------------------------<
%  Procedure : printLst/1
%  Purpose   : Prints ths name of the peephole optimizations done in the
%              current function being compiled.
%  Arguments : Lst - A list of numbers.
%  Return    : unit
%  Notes     : Prints (append) the atoms in the list with a space in 
%              between to a file called 'peepcount.txt'. 
%%>----------------------------------------------------------------------<
printLst(Lst) ->
    case ?DO_LOGGING of
	true ->
	    {ok, Dir} = file:get_cwd(),
	    file:set_cwd(?LOG_DIR),
	    {ok, File} = file:open(?LOG_FILE, [read, append]),
	    printLst(File, Lst),
	    file:set_cwd(Dir),
	    file:close(File);
	_ ->
	    ok
    end.
printLst(File, [Opt|Lst]) ->
   %io:format(File, "Peephole applied ~w times!\n", [length(Lst)]),
    io:format(File, "Peephole optimization applied: ", []),
    io:write(File, Opt),
    io:format(File, "\n", []),
    printLst(File, Lst);
printLst(_, []) -> done.



%%  MISC CODE!!!
%% >-------------------------------------------------------------------------< 

%% Log2 function
%% -------------
%% Used by ElimMDPow2 clause of peep(..)
% log2(Nr) -> log2(Nr, 0).
% log2(0, _) -> none;
% log2(Nr, I) ->
%     case (Nr band 1) == 1 of
% 	true ->
% 	    case Nr of
% 		1 ->
% 		    {true, I, positive};
% 		-1 ->
% 		    {true, I, negative};
% 		_ ->
% 		    {false, 0, positive}
% 	    end;
% 	false ->
% 	    log2((Nr bsr 1), I+1)
%     end.

%% Skips through all comments and move instructions and returns the next one
%% ------------------------------------------------------------------------
%% Used by ElimCmpTest above.
check([I|Ins]) ->
    case I of
	#comment{} ->
	    check(Ins);
	#move{} ->
	    check(Ins);
	#fmov{} ->
	    check(Ins);
	#movsx{} ->
	    check(Ins);
	#movzx{} ->
	    check(Ins);
	OtherI ->
	    OtherI
    end.


%%% Replacing add/sub 1 with inc/dec is incorrect if the carry flag
%%% defined by the add/sub is used later -- disabled for now.
%peep([#alu{aluop='sub',src=#x86_imm{value=1},dst=Dst} | Insns]) ->
%    [hipe_x86:mk_dec(Dst) | peep(Insns)];
%peep([#alu{aluop='add',src=#x86_imm{value=1},dst=Dst} | Insns]) ->
%    [hipe_x86:mk_inc(Dst) | peep(Insns)];
%peep([#push{src=Src} | (Insns = [#pop{src=Src} | _])]) -> peep(Insns);


%% Old Working Code - Left for safety
%% --------------------------
% peepN(Insns) -> peepN(Insns, []).    
% peepN([#jmp_label{label=Label} | (Insns = [#label{label=Label}|_])], Lst) ->
%     peepN(Insns, [nearGotoElim | Lst]);
% peepN([I | Insns], Lst) ->
%     [I | peepN(Insns, Lst)];
% peepN([], Lst) ->
%     printLst(Lst),
%     [].

