%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose: Implements an "MG" used by the test suite
%%----------------------------------------------------------------------
-module(megaco_test_mg).

-export([start/5, start/6, stop/1, stop/2, 
	 get_stats/1, get_stats/2, reset_stats/1, reset_stats/2,
	 user_info/1, user_info/2, user_info/3, 
	 update_user_info/3, update_user_info/4, 
	 conn_info/1, conn_info/2, conn_info/3, 
	 update_conn_info/3, update_conn_info/4, 
	 service_change/1, service_change/2, 
	 ack_info/2, ack_info/3, 
	 group_requests/2, group_requests/3, 
	 notify_request/1, notify_request/2, 
	 await_notify_reply/1, 
	 notify_request_and_reply/1, notify_request_and_reply/2, 
	 cancel_request/2, cancel_request/3, 
	 apply_load/2, apply_load/3, 
	 apply_multi_load/3, apply_multi_load/4, 
	 enable_test_code/4, 
	 verbosity/2]).

-export([mg/3, notify_request_handler_main/5]).
-export([loader_main/3]).

%% Megaco callback api
-export([
	 handle_connect/4,
	 handle_disconnect/5,
	 handle_syntax_error/5,
	 handle_message_error/5,
	 handle_trans_request/5,
	 handle_trans_long_request/5,
	 handle_trans_reply/6,
	 handle_trans_ack/6
	]).

-include("megaco_test_lib.hrl").
-include_lib("megaco/include/megaco.hrl").
-include_lib("megaco/include/megaco_message_v1.hrl").

-define(A4444, tid(255*256*256)                 ).
-define(A4445, tid(255*256*256 + 255)           ).
-define(A5555, tid(255*256*256 + 255*256)       ).
-define(A5556, tid(255*256*256 + 255*256 + 255) ).

-record(mg, {mid           = undefined,
	     state         = initiated,
	     req_handler   = undefined,
	     call_mode     = async,
	     group_size    = 1,
	     ack_info      = undefined,
	     load_counter  = 0,
	     reply_counter = 0,
	     mload_info    = undefined}).


%%% --------------------------------------------------------------------

start(Node, Mid, Encoding, Transport, Verbosity) ->
    start(Node, Mid, Encoding, Transport, [], Verbosity).

start(Node, Mid, Encoding, Transport, Conf, Verbosity) ->
    d("start mg[~p]: ~p", [Node, Mid]),
    RI1 = encoding_config(Encoding),
    RI2 = transport_config(Transport),
    RI = {receive_info, RI1 ++ RI2},
    Config = [{local_mid, Mid}, RI] ++ Conf,
    Pid = spawn_link(Node, ?MODULE, mg, [self(), Verbosity, Config]),
    await_started(Pid).

encoding_config(text) ->
    [{encoding_module, megaco_pretty_text_encoder},
     {encoding_config, []},
     {port,2944}];
encoding_config(binary) ->
    [{encoding_module, megaco_ber_bin_encoder},
     {encoding_config, []},
     {port,2945}].

transport_config(tcp) ->
    [{transport_module, megaco_tcp}];
transport_config(udp) ->
    [{transport_module, megaco_udp}].

await_started(Pid) ->
    receive
	{started, Pid} ->
	    d("await_started ~p: ok", [Pid]),
	    {ok, Pid};
	{'EXIT', Pid, Reason} ->
	    i("await_started ~p: received exit signal: ~p", [Pid, Reason]),
	    exit({failed_starting, Pid, Reason})
    after 10000 ->
	    i("await_started ~p: timeout", [Pid]),
	    exit({error, timeout})
    end.


verbosity(Pid, V) ->
    Pid ! {verbosity, V, self()}.


stop(Pid) ->
    server_request(Pid, stop, stopped).

stop(Pid, Mid) ->
    server_request(Pid, {stop, Mid}, {stopped, Mid}).


get_stats(Pid) ->
    server_request(Pid, statistics, statistics_reply).

get_stats(Pid, Mid) ->
    server_request(Pid, {statistics, Mid}, {statistics_reply, Mid}).


reset_stats(Pid) ->
    server_request(Pid, reset_stats, reset_stats_ack).

reset_stats(Pid, Mid) ->
    server_request(Pid, {reset_stats, Mid}, reset_stats_ack).


user_info(Pid) ->
    server_request(Pid, {user_info, all}, user_info_ack).

user_info(Pid, Tag) when atom(Tag) ->
    server_request(Pid, {user_info, Tag}, user_info_ack);
user_info(Pid, Mid) ->
    server_request(Pid, {user_info, Mid, all}, user_info_ack).

user_info(Pid, Mid, Tag) when atom(Tag) ->
    server_request(Pid, {user_info, Mid, Tag}, user_info_ack).


update_user_info(Pid, Tag, Val) ->
    server_request(Pid, {update_user_info, Tag, Val}, update_user_info_ack).

update_user_info(Pid, Mid, Tag, Val) ->
    server_request(Pid, {update_user_info, Mid, Tag, Val}, update_user_info_ack).


conn_info(Pid) ->
    server_request(Pid, {conn_info, all}, conn_info_ack).

conn_info(Pid, Tag) when atom(Tag) ->
    server_request(Pid, {conn_info, Tag}, conn_info_ack);
conn_info(Pid, Mid) ->
    server_request(Pid, {conn_info, Mid, all}, conn_info_ack).

conn_info(Pid, Mid, Tag) ->
    server_request(Pid, {conn_info, Mid, Tag}, conn_info_ack).


update_conn_info(Pid, Tag, Val) ->
    server_request(Pid, {update_conn_info, Tag, Val}, update_conn_info_ack).

update_conn_info(Pid, Mid, Tag, Val) ->
    server_request(Pid, {update_conn_info, Mid, Tag, Val}, update_conn_info_ack).


enable_test_code(Pid, Module, Where, Fun) 
  when atom(Module), atom(Where), function(Fun) ->
    Tag = {Module, Where},
    server_request(Pid, {enable_test_code, Tag, Fun}, enable_test_code_reply).

service_change(Pid) ->
    server_request(Pid, service_change, service_change_reply).

service_change(Pid, Mid) ->
    server_request(Pid, {service_change, Mid}, service_change_reply).


group_requests(Pid, N) ->
    server_request(Pid, {group_requests, N}, group_requests_reply).

group_requests(Pid, Mid, N) ->
    server_request(Pid, {group_requests, Mid, N}, group_requests_reply).


ack_info(Pid, InfoPid) ->
    Pid ! {{ack_info, InfoPid}, self()}.

ack_info(Pid, Mid, InfoPid) ->
    Pid ! {{ack_info, Mid, InfoPid}, self()}.


notify_request(Pid) ->
    Pid ! {notify_request, self()}.

notify_request(Pid, Mid) ->
    Pid ! {{notify_request, Mid}, self()}.


await_notify_reply(Pid) ->
    await_reply(Pid, notify_request_reply).


notify_request_and_reply(Pid) ->
    notify_request(Pid),
    await_notify_reply(Pid).

notify_request_and_reply(Pid, Mid) ->
    notify_request(Pid, Mid),
    await_notify_reply(Pid).


cancel_request(Pid, Reason) ->
    server_request(Pid, cancel_request, Reason, cancel_request_reply).

cancel_request(Pid, Mid, Reason) ->
    server_request(Pid, {cancel_request, Mid}, Reason, cancel_request_reply).


apply_load(Pid, CounterStart) ->
    server_request(Pid, apply_load, CounterStart, apply_load_ack).

apply_load(Pid, Mid, CounterStart) ->
    server_request(Pid, {apply_load, Mid}, CounterStart, apply_load_ack).


apply_multi_load(Pid, NumLoaders, NumReq) ->
    server_request(Pid, apply_multi_load, {NumLoaders, NumReq}, apply_multi_load_ack).

apply_multi_load(Pid, Mid, NumLoaders, NumReq) ->
    server_request(Pid, {apply_multi_load, Mid}, {NumLoaders, NumReq}, apply_multi_load_ack).


server_request(Pid, Req, ReplyTag) ->
    Pid ! {Req, self()},
    await_reply(Pid, ReplyTag).

server_request(Pid, Req, ReqData, ReplyTag) ->
    Pid ! {Req, ReqData, self()},
    await_reply(Pid, ReplyTag).

await_reply(Pid, ReplyTag) ->
    await_reply(Pid, ReplyTag, infinity).

await_reply(Pid, ReplyTag, Timeout) ->
    receive
	{ReplyTag, Reply, Pid} ->
	    Reply;
	{'EXIT', Pid, Reason} ->
	    exit({failed, ReplyTag, Pid, Reason})
    after Timeout ->
	    exit({timeout, ReplyTag, Pid})
    end.


server_reply(Pid, ReplyTag, Reply) ->
    Pid ! {ReplyTag, Reply, self()}.


%%% --------------------------------------------------------------------


mg(Parent, Verbosity, Config) ->
    process_flag(trap_exit, true),
    put(verbosity, Verbosity),
    put(sname,   "MG"),
    i("mg -> starting"),
    Mid = init(Config),
    notify_started(Parent),
    DefaultMG = #mg{mid = Mid},
    ets:insert(megaco_test_mg, DefaultMG),
    i("mg -> started"),
    case (catch loop(Parent)) of
	{'EXIT', normal} ->
	    exit(normal);
	{'EXIT', Reason} ->
	    i("mg failed with reason:~n   ~p", [Reason]),
	    exit(Reason);
	Else ->
	    i("mg terminated: ~n   ~p", [Else]),
	    exit({unexpected, Else})
    end.

init(Config) ->
    d("init -> entry with"
      "~n   Config: ~p", [Config]),
    random_init(),
    ets:new(megaco_test_mg, [public, named_table, {keypos, 2}]),
    Mid = get_conf(local_mid, Config),
    RI  = get_conf(receive_info, Config),
    d("init -> start megaco"),
    application:start(megaco),

    d("init -> start megaco user"),
    Conf0 = lists:keydelete(local_mid,    1, Config),
    Conf1 = lists:keydelete(receive_info, 1, Conf0),
    ok = megaco:start_user(Mid, Conf1),
    d("init -> update user info (user_mod)"),
    ok = megaco:update_user_info(Mid, user_mod,  ?MODULE),
    d("init -> update user info (user_args)"),
    ok = megaco:update_user_info(Mid, user_args, [self(), Mid]),

%     d("init -> start megaco user"),
%     megaco:start_user(Mid, []),
%     d("init -> update user info (user_mod)"),
%     megaco:update_user_info(Mid, user_mod,  ?MODULE),
%     d("init -> update user info (user_args)"),
%     megaco:update_user_info(Mid, user_args, [self()]),

    d("init -> get user info (receive_handle)"),
    RH = megaco:user_info(Mid,receive_handle),
    d("init -> parse receive info"),
    {MgcPort,RH1} = parse_receive_info(RI, RH),
    d("init -> start transport"),
    {ok, _CH} = start_transport(MgcPort, RH1),
    Mid.


loop(Parent) ->
    d("loop -> await request", []),
    receive
	{verbosity, V, Parent} ->
	    i("loop -> received new verbosity: ~p", [V]),
	    put(verbosity,V),
	    loop(Parent);


	{stop, Parent} ->
	    i("loop -> stopping", []),
	    Res = do_stop(),
	    d("loop -> stop result: ~p", [Res]),
	    server_reply(Parent, stopped, {ok, Res}),
	    exit(normal);

	{{stop, Mid}, Parent} ->
	    i("loop -> stop user ~p", [Mid]),
	    Res = do_stop(Mid),
	    server_reply(Parent, {stopped, Mid}, {ok, Res}),
	    loop(Parent);

	
	{{enable_test_code, Tag, Fun}, Parent} ->
	    i("loop -> enable_test_code: ~p, ~p", [Tag, Fun]),
	    Reply = (catch ets:insert(megaco_test_data, {Tag, Fun})),
	    d("loop -> enable_test_code -> "
	      "~n   Reply:                          ~p"
	      "~n   ets:tab2list(megaco_test_data): ~p", 
	      [Reply,ets:tab2list(megaco_test_data)]),
	    server_reply(Parent, enable_test_code_reply, Reply),
	    loop(Parent);

	%% Give me statistics
	{statistics, Parent} ->
	    i("loop -> got request for statistics", []),
	    Stats = do_get_statistics(),
	    server_reply(Parent, statistics_reply, {ok, Stats}),
	    loop(Parent); 

	{{statistics, Mid}, Parent} ->
	    i("loop -> got request for statistics", []),
	    Stats = do_get_statistics(Mid),
	    server_reply(Parent, {statistics_reply, Mid}, {ok, Stats}),
	    loop(Parent); 


	{reset_stats, Parent} ->
	    i("loop -> got request to reset stats counters", []),
	    do_reset_stats(),
	    server_reply(Parent, reset_stats_ack, ok),
	    loop(Parent);

	{{reset_stats, Mid}, Parent} ->
	    i("loop -> got request to reset stats counters", []),
	    do_reset_stats(Mid),
	    server_reply(Parent, {reset_stats_ack, Mid}, ok),
	    loop(Parent);


	{{user_info, Tag}, Parent} ->
	    i("loop -> got user_info request for ~w", [Tag]),
	    Res = do_get_user_info(Tag),
	    d("loop -> Res: ~p", [Res]),
	    server_reply(Parent, user_info_ack, Res),
	    loop(Parent);

	{{user_info, Mid, Tag}, Parent} ->
	    i("loop -> [~p] got user_info request for ~w", [Mid, Tag]),
	    Res = do_get_user_info(Mid, Tag),
	    d("loop -> Res: ~p", [Res]),
	    server_reply(Parent, {user_info_ack, Mid}, Res),
	    loop(Parent);


	{{update_user_info, Tag, Val}, Parent} ->
	    i("loop -> got update_user_info: ~w -> ~p", [Tag, Val]),
	    Res = do_update_user_info(Tag, Val),
	    d("loop -> Res: ~p", [Res]),
	    server_reply(Parent, update_user_info_ack, Res),
	    loop(Parent);

	{{update_user_info, Mid, Tag, Val}, Parent} ->
	    i("loop -> [~p] got update_user_info: ~w -> ~p", [Mid, Tag, Val]),
	    Res = do_update_user_info(Mid, Tag, Val),
	    d("loop -> Res: ~p", [Res]),
	    server_reply(Parent, {update_user_info_ack, Mid}, Res),
	    loop(Parent);


	{{conn_info, Tag}, Parent} ->
	    i("loop -> got conn_info request for ~w", [Tag]),
	    Res = do_get_conn_info(Tag),
	    server_reply(Parent, conn_info_ack, Res),
	    loop(Parent);

	{{conn_info, Mid, Tag}, Parent} ->
	    i("loop -> got conn_info request for ~w", [Tag]),
	    Res = do_get_conn_info(Mid, Tag),
	    server_reply(Parent, {conn_info_ack, Mid}, Res),
	    loop(Parent);


	{{update_conn_info, Tag, Val}, Parent} ->
	    i("loop -> got update_conn_info: ~w -> ~p", [Tag, Val]),
	    Res = do_update_conn_info(Tag, Val),
	    server_reply(Parent, update_conn_info_ack, Res),
	    loop(Parent);

	{{update_conn_info, Mid, Tag, Val}, Parent} ->
	    i("loop -> got update_conn_info: ~w -> ~p", [Tag, Val]),
	    Res = do_update_conn_info(Mid, Tag, Val),
	    server_reply(Parent, {update_conn_info_ack, Mid}, Res),
	    loop(Parent);


	%% Do a service change
	%% No server-reply here. Since the service change is 
	%% async, the reply (from the MGC) will come later.
	{service_change, Parent} ->
	    i("loop -> received request to perform service change", []),
	    case do_service_change() of
		ok ->
		    d("loop -> service change started", []);
		Error ->
		    d("loop -> service change failed: ~p", [Error]),
		    server_reply(Parent, service_change_reply, Error)
	    end,
	    loop(Parent); 

	{{service_change, Mid}, Parent} ->
	    i("loop -> [~p] received request to perform service change", [Mid]),
	    case do_service_change(Mid) of
		ok ->
		    d("loop -> [~p] service change started", [Mid]);
		Error ->
		    d("loop -> [~p] service change failed: ~p", [Mid, Error]),
		    server_reply(Parent, {service_change_reply, Mid}, Error)
	    end,
	    loop(Parent); 


	{{group_requests, N}, Parent} when N > 0 ->
	    i("loop -> received group_requests ~p", [N]),
	    Reply = do_update_group_requests(N),
	    server_reply(Parent, group_requests_reply, Reply),
	    loop(Parent); 

	{{group_requests, Mid, N}, Parent} when N > 0 ->
	    i("loop -> received group_requests ~p", [N]),
	    Reply = do_update_group_requests(Mid, N),
	    server_reply(Parent, {group_requests_reply, Mid}, Reply),
	    loop(Parent); 


	{{ack_info, To}, Parent} ->
	    i("loop -> received request to inform about received ack's ", []),
	    do_update_ack_info(To),
	    loop(Parent);

	{{ack_info, Mid, To}, Parent} ->
	    i("loop -> [~p] received request to inform about received ack's ", [Mid]),
	    do_update_ack_info(Mid, To),
	    loop(Parent);


	%% Make a sync-call
	{notify_request, Parent} ->
	    i("loop -> received request to send notify request ", []),
	    Res = do_handle_notify_request(),
	    d("loop -> notify request result: ~p", [Res]),
	    loop(Parent);

	{{notify_request, Mid}, Parent} ->
	    i("loop -> [~p] received request to send notify request ", [Mid]),
	    Res = do_handle_notify_request(Mid),
	    d("loop -> notify request result: ~p", [Res]),
	    loop(Parent);


	%% sync-call complete
	{notify_request_complete, Mid, NotifyReply, Pid} ->
	    i("loop -> [~p] received notify request complete from "
	      "~n   ~p with"
	      "~n   NotifyReply: ~p", 
	      [Mid, Pid, NotifyReply]),
	    Reply = do_handle_notify_request_complete(Mid, NotifyReply, Pid),
	    d("loop -> Reply: ~p", [Reply]),
	    server_reply(Parent, notify_request_reply, Reply),
	    loop(Parent);


	%% cancel requests
	{cancel_request, Reason, Parent} ->
	    i("loop -> received request to cancel (all) megaco requests ", []),
	    Res = do_cancel_requests(Reason),
	    server_reply(Parent, cancel_request_reply, Res),
	    loop(Parent);

	{{cancel_request, Mid}, Reason, Parent} ->
	    i("loop -> received request to cancel (all) megaco requests ", []),
	    Res = do_cancel_requests(Mid, Reason),
	    server_reply(Parent, cancel_request_reply, Res),
	    loop(Parent);


	%% Apply multi-load
	{apply_multi_load, {NL, NR}, Parent} -> 
	    i("loop -> received apply_multi_load request", []),
	    ok = start_loaders(NL, NR),
	    loop(Parent);

	{{apply_multi_load, Mid}, {NL, NR}, Parent} -> 
	    i("loop -> received apply_multi_load request", []),
	    ok = start_user_loaders(Mid, NL, NR),
	    loop(Parent);


	%% Apply some load
	{{apply_load, Mid}, Times, Parent} ->
	    i("loop -> received apply_load request", []),
	    apply_load_timer(Mid),
	    server_reply(Parent, apply_load_ack, ok),
	    loop(Parent);


	{apply_load_timeout, Mid} ->
	    d("loop -> received apply_load timeout ~p", [Mid]),
	    do_apply_load(Parent, Mid),
	    loop(Parent);


	%% Megaco callback messages
	{request, Request, Mid, From} ->
	    d("loop -> received megaco request: ~n   ~p"
	      "~n   Mid:  ~p"
	      "~n   From: ~p", 
	      [Request, Mid, From]),
	    Reply = handle_megaco_request(Request, Parent, Mid),
	    d("loop -> send (megaco callback) request reply: ~n~p", [Reply]),
	    From ! {reply, Reply, self()},
	    loop(Parent);


	{'EXIT', Pid, Reason} -> 
	    i("loop -> received exit signal:"
	      "~n   Pid: ~p"
	      "~n   Mid: ~p", [Pid, Reason]),
	    handle_exit(Parent, Pid, Reason),
	    loop(Parent);


	Invalid ->
	    error_msg("received invalid request: ~n~p", [Invalid]),
	    loop(Parent)

    end.

%%
%% Stop all or some user (mid)
%%
do_stop() ->
    [do_stop(Mid) || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)].

do_stop(Mid) ->
    d("do_stop -> stopping user ~p", [Mid]),
    Disco = fun close_conn/1, 
    lists:map(Disco, megaco:user_info(Mid, connections)),
    megaco:stop_user(Mid),
    ets:delete(megaco_test_mg, Mid).

close_conn(CH) ->
    d("do_stop -> closing connection ~p", [CH]),
    Reason     = {stopped_by_user,self()},
    Pid        = megaco:conn_info(CH, control_pid),
    SendMod    = megaco:conn_info(CH, send_mod),
    SendHandle = megaco:conn_info(CH, send_handle),
    megaco:disconnect(CH, Reason),
    megaco:cancel(CH, Reason),
    case SendMod of
	megaco_tcp -> megaco_tcp:close(SendHandle);
	megaco_udp -> megaco_udp:close(SendHandle);
	SendMod    -> exit(Pid, Reason)
    end.



%%
%% reset all or some user stats
%%
do_reset_stats() ->
    Mids = [Mid || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)],
    lists:foreach(fun do_reset_stats/1, Mids).

do_reset_stats(Mid) ->
    %% We only have one connection
    [CH] = megaco:user_info(Mid, connections),
    do_reset_stats1(CH).

do_reset_stats1(CH) ->
    megaco:reset_stats(),
    case (catch megaco:conn_info(CH, send_mod)) of
	{error, Reason} ->
	    error_msg("unexpected result when retrieving send module for "
		      "own connection ~p: ~p. "
		      "~nexiting...", [CH, Reason]),
	    exit({invalid_connection, CH, Reason});
	{'EXIT', Reason} ->
	    error_msg("exit signal when retrieving send module for "
		      "own connection ~p: ~p. "
		      "~nexiting...", [CH, Reason]),
	    exit({invalid_connection, CH, Reason});
	SendMod when atom(SendMod) ->
	    SendMod:reset_stats()
    end.


%%
%% Update user info for some or all user(s)
%%
do_update_user_info(Tag, Val) ->
    Mids = [Mid || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)],
    [do_update_user_info(Mid, Tag, Val) || Mid <- Mids].

do_update_user_info(Mid, Tag, Val) ->
    (catch megaco:update_user_info(Mid, Tag, Val)).


%%
%% Get user info for some or all user(s)
%%
do_get_user_info(Tag) ->
    Mids = [Mid || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)],
    [{Mid, do_get_user_info(Mid, Tag)} || Mid <- Mids].

do_get_user_info(Mid, Tag) ->
    (catch megaco:user_info(Mid, Tag)).


%%
%% Update conn info for some or all user(s)
%%
do_update_conn_info(Tag, Val) ->
    Mids = [Mid || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)],
    [do_update_conn_info(Mid, Tag, Val) || Mid <- Mids].

do_update_conn_info(Mid, Tag, Val) ->
    %% We only have one connection
    [CH] = megaco:user_info(Mid, connections),
    do_update_conn_info1(CH, Tag, Val).

do_update_conn_info1(CH, Tag, Val) ->
    (catch megaco:update_conn_info(CH, Tag, Val)).


%%
%% Get conn info for some or all user(s)
%%
do_get_conn_info(Tag) ->
    Mids = [Mid || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)],
    [{Mid, do_get_conn_info(Mid, Tag)} || Mid <- Mids].

do_get_conn_info(Mid, Tag) ->
    Conns = megaco:user_info(Mid, connections),
    [{CH, (catch megaco:user_info(Mid, Tag))} || CH <- Conns].


%% 
%% Get statistics for some or all user(s)
%% 
do_get_statistics() ->
    Mids = [Mid || #mg{mid = Mid} <- ets:tab2list(megaco_test_mg)],
    [{Mid, do_get_statistics(Mid)} || Mid <- Mids].

do_get_statistics(Mid) ->
    %% We only have one connection
    [CH] = megaco:user_info(Mid, connections),
    do_get_conn_statistics(CH).

do_get_conn_statistics(CH) ->
    {ok, Gen}  = megaco:get_stats(),
    Pid        = megaco:conn_info(CH, control_pid),
    SendMod    = megaco:conn_info(CH, send_mod),
    SendHandle = megaco:conn_info(CH, send_handle),
    {ok, Trans} = 
	case SendMod of
	    megaco_tcp -> megaco_tcp:get_stats(SendHandle);
	    megaco_udp -> megaco_udp:get_stats(SendHandle)
	end,
    [{gen, Gen}, {trans, Trans}].
    

%%
%% Perform service change for some or all user(s)
%%
do_service_change() ->
    d("start service changes for all MGs"),
    MGs = [MG || MG <- ets:tab2list(megaco_test_mg)],
    lists:foreach(fun do_service_change/1, MGs).

do_service_change(#mg{mid = Mid, state = initiated} = MG) ->
    %% We only have one connection
    d("do service change for ~p", [Mid]),
    [CH]   = megaco:user_info(Mid, connections),
    Method = restart,
    Reason = ?megaco_cold_boot,
    case do_service_change(CH, Method, Reason) of
	ok ->
	    ets:insert(megaco_test_mg, MG#mg{state = connecting}),
	    ok;
	Error ->
	    Error
    end;
do_service_change(MG) when record(MG, mg) ->
    {error, invalid_state};
do_service_change(Mid) ->
    case ets:lookup(megaco_test_mg, Mid) of
	[MG] ->
	    do_service_change(MG);
	_ ->
	    {error, not_found}
    end.

do_service_change(ConnHandle, Method, Reason) ->
    d("sending service change using:"
      "~n   ConnHandle: ~p"
      "~n   Method:     ~p"
      "~n   Reason:     ~p", [ConnHandle, Method, Reason]),
    SCP    = cre_serviceChangeParm(Method, [Reason]),
    TermId = [?megaco_root_termination_id],
    SCR    = cre_serviceChangeReq(TermId, SCP),
    CR     = cre_commandReq({serviceChangeReq, SCR}),
    AR     = cre_actionReq(?megaco_null_context_id,[CR]),
    megaco:cast(ConnHandle, [AR], []).


%%
%% update group requests
%%
do_update_group_requests(N) ->
    First = ets:first(megaco_test_mg),
    do_update_group_requests(N, First, []).

do_update_group_requests(N, '$end_of_table', Acc) ->
    {ok, lists:reverse(Acc)};
do_update_group_requests(N, Mid, Acc) ->
    case do_update_group_requests(N, Mid) of
	{ok, OldN} ->
	    Next = ets:next(megaco_test_mg, Mid), 
	    do_update_group_requests(N, Next, [{Mid, OldN}|Acc]);
	Error ->
	    Error
    end.

do_update_group_requests(N, Mid) ->
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{group_size = OldN} = MG] ->
	    ets:insert(megaco_test_mg, MG#mg{group_size = N}),
	    {ok, OldN};
	_ ->
	    {error, {not_found, Mid}}
    end.
    

%%
%% Update ack_info for some or all user(s)
%%
do_update_ack_info(To) ->
    First = ets:first(megaco_test_mg),
    do_update_ack_info(To, First).

do_update_ack_info(To, '$end_of_table') ->
    ok;
do_update_ack_info(To, Mid) ->
    case do_update_ack_info1(To, Mid) of
	ok ->
	    do_update_ack_info(To, ets:next(megaco_test_mg, Mid));
	Error ->
	    Error
    end.

do_update_ack_info1(To, Mid) ->
    case ets:lookup(megaco_test_mg, Mid) of
	[MG] ->
	    ets:insert(megaco_test_mg, MG#mg{ack_info = To}),
	    ok;
	_ ->
	    {error, {not_found, Mid}}
    end.


%% Make a sync call for some or all user(s)
do_handle_notify_request() ->
    First = ets:first(megaco_test_mg),
    do_handle_notify_request1(First).

do_handle_notify_request1('$end_of_table') ->
    ok;
do_handle_notify_request1(Mid) ->
    case do_handle_notify_request(Mid) of
	ok ->
	    do_handle_notify_request1(ets:next(megaco_test_mg, Mid));
	Error ->
	    Error
    end.

do_handle_notify_request(Mid) ->
    d("do_handle_notify_request -> entry with"
      "~n   Mid: ~p", [Mid]),
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{group_size = N, state = connected} = MG] ->
	    [CH] = megaco:user_info(Mid, connections),
	    Pid = start_notify_request_handler(Mid, CH, N),
	    ets:insert(megaco_test_mg, MG#mg{req_handler = Pid}),
	    ok;
	[#mg{state = State}] ->
	    {error, {invalid_state, State}};
	_ ->
	    {error, {not_found, Mid}}
    end.

do_handle_notify_request_complete(Mid, NotifyReply, Pid) ->
    case ets:lookup(megaco_test_mg, Mid) of
	[MG] ->
	    ets:insert(megaco_test_mg, MG#mg{req_handler = undefined}),
	    NotifyReply;
	_ ->
	    {error, {not_found, Mid, NotifyReply}}
    end.



%%
%% Cancel requests for some or all user(s)
%%
do_cancel_requests(Reason) ->
    First = ets:first(megaco_test_mg),
    do_cancel_requests(Reason, First, []).

do_cancel_requests(_Reason, '$end_of_table', Acc) ->
    {ok, lists:reverse(Acc)};
do_cancel_requests(Reason, Mid, Acc) ->
    Res = do_cancel_requests(Mid, Reason),
    do_cancel_requests(Reason, ets:next(megaco_test_mg, Mid), [{Mid,Res}|Acc]).

do_cancel_requests(Mid, Reason) ->
    [CH] = megaco:user_info(Mid, connections),
    megaco:cancel(CH, Reason).    


%% 
%% Apply load for some or all user(s)
%% 
start_loaders(NumLoaders, Times) ->
    First = ets:first(megaco_test_mg),
    start_loaders(First, NumLoaders, Times).

start_loaders('$end_of_table', _NumLoaders, _Times) ->
    ok;
start_loaders(Mid, NumLoaders, Times) ->
    case start_user_loaders(Mid, NumLoaders, Times) of
	ok ->
	    start_loaders(ets:next(megaco_test_mg, Mid), NumLoaders, Times);
	Error ->
	    Error
    end.

start_user_loaders(Mid, NumLoaders, Times) ->
    [CH] = megaco:user_info(Mid, connections),
    Loaders = start_user_loaders1(NumLoaders, [], [Mid, Times, CH]),
    case ets:lookup(megaco_test_mg, Mid) of
	[MG] ->
	    ets:insert(megaco_test_mg, MG#mg{mload_info = {Loaders, 0, 0}}),
	    ok;
	_ ->
	    {error, {not_found, Mid}}
    end.

start_user_loaders1(0, Acc, _) ->
    Acc;
start_user_loaders1(N, Acc, Args) ->
    Pid = spawn_link(?MODULE, loader_main, Args),
    start_user_loaders1(N-1, [Pid|Acc], Args).

loader_main(Mid, 0, _) ->
    exit({loader_done, Mid});
loader_main(Mid, N, CH) ->
    {Act, _} = make_notify_request(),
    Res      = megaco:call(CH, Act, []),
    loader_main(Mid, N-1, CH).



handle_exit(Parent, Parent, Reason) ->
    exit({parent_terminated, Reason});

handle_exit(Parent, Pid, {loader_done, Mid}) ->
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{mload_info = {Loaders0, Ok, Err}} = MG] ->
	    Loaders = lists:delete(Pid, Loaders0),
	    LoadInfo =
		case Loaders of
		    [] ->
			d("handle_exit -> multi load done", []),
			server_reply(Parent, apply_multi_load_ack, {ok, Ok+1, Err}),
			undefined;
		    _ ->
			{Loaders, Ok+1, Err}
		end,
	    ets:insert(megaco_test_mg, MG#mg{mload_info = LoadInfo}),
	    ok;
	[] ->
	    i("loader_done [~p] mid ~p not found", [Pid, Mid]),
	    ok
    end;

handle_exit(Parent, Pid, Reason) ->
    Pat = {'_', 
	   #mg{mid           = '_', 
	       state         = '_', 
	       req_handler   = Pid,
	       call_mode     = '_',
	       group_size    = '_',
	       ack_info      = '_',
	       load_counter  = '_',
	       reply_counter = '_',
	       mload_info    = '_'}
	  },
    case ets:match_object(megaco_test_mg, Pat) of
	[{Mid, MG}] ->
	    error_msg("MG ~p "
		      "received unexpected exit from the request handler:"
		      "~n   ~p", [Mid, Reason]),
	    server_reply(Parent, notify_request_reply, 
			 {error, {request_handler_exit, Reason}}),
	    ets:insert(megaco_test_mg, MG#mg{req_handler = undefined}),
	    ok;
	[] ->
	    MGs = ets:tab2list(megaco_test_mg),
	    case update_loader(Parent, MGs, Pid) of
		not_found ->
		    error_msg("received unexpected exit signal from ~p:~n~p", 
			      [Pid, Reason]),
		    ok;
		_ ->
		    ok
	    end
    end.

update_loader(_, [], _) ->
    not_found;
update_loader(Parent, [#mg{mload_info = {Loaders, Ok, Err}} = MG|MGs], Pid) 
  when length(Loaders) > 0 ->
    case lists:delete(Pid, Loaders) of
	[] -> 
	    %% since we cannot be empty prior the delete, 
	    %% the last one exited...
	    server_reply(Parent, apply_multi_load, {ok, Ok, Err+1}),
	    ets:insert(megaco_test_mg, MG#mg{mload_info = undefined}),
	    done;
	Loaders ->
	    %% Could not be this MG, so go on to the next
	    update_loader(Parent, MGs, Pid);
	Loaders1 ->
	    %% Not empty, but we removed one
	    ets:insert(megaco_test_mg, 
		       MG#mg{mload_info = {Loaders1,Ok,Err+1}}),
	    updated
    end;
update_loader(Parent, [_|MGs], Pid) ->
    update_loader(Parent, MGs, Pid).

		      
parse_receive_info(RI, RH) ->
    d("parse_receive_info -> get encoding module"),
    EM = get_encoding_module(RI),
    d("parse_receive_info -> get encoding config"),
    EC = get_encoding_config(RI, EM),
    d("parse_receive_info -> get transport module"),
    TM = get_transport_module(RI),
    d("parse_receive_info -> get transport port"),
    TP = get_transport_port(RI),
    RH1 = RH#megaco_receive_handle{send_mod        = TM,
				   encoding_mod    = EM,
				   encoding_config = EC},
    {TP, RH1}.


start_transport(MgcPort,
		   #megaco_receive_handle{send_mod = megaco_tcp} = RH) ->
    start_tcp(MgcPort,RH);
start_transport(MgcPort,
		   #megaco_receive_handle{send_mod = megaco_udp} = RH) ->
    start_udp(MgcPort,RH);
start_transport(_, #megaco_receive_handle{send_mod = Mod}) ->
    throw({error, {bad_send_mod, Mod}}).


start_tcp(MgcPort, RH) ->
    d("start tcp transport (~p)", [MgcPort]),
    case megaco_tcp:start_transport() of
	{ok, Sup} ->
	    d("tcp transport started: ~p", [Sup]),
	    {ok, LocalHost} = inet:gethostname(),
	    Opts = [{host, LocalHost},{port, MgcPort}, {receive_handle, RH}],
	    d("tcp connect", []),
	    case megaco_tcp:connect(Sup, Opts) of
		{ok, SendHandle, ControlPid} ->
		    d("tcp connected: ~p, ~p", [SendHandle, ControlPid]),
		    megaco_tcp_connect(RH, LocalHost, SendHandle, ControlPid);
		{error, Reason} ->
		    throw({error, {megaco_tcp_connect, Reason}})
	    end;
        {error, Reason} ->
            throw({error, {megaco_tcp_start_transport, Reason}})
    end.

megaco_tcp_connect(RH, LocalHost, SendHandle, ControlPid) ->
    PrelMgcMid = preliminary_mid,
    d("megaco connect", []),    
    {ok, CH} = megaco:connect(RH, PrelMgcMid, SendHandle, ControlPid),
    d("megaco connected: ~p", [CH]),    
    {ok, CH}.

start_udp(MgcPort, RH) ->
    d("start udp transport (~p)", [MgcPort]),
    case megaco_udp:start_transport() of
	{ok, Sup} ->
	    d("udp transport started: ~p", [Sup]),
	    Opts = [{port, 0}, {receive_handle, RH}],
	    d("udp open", []),
	    case megaco_udp:open(Sup, Opts) of
		{ok, Handle, ControlPid} ->
		    d("udp opened: ~p, ~p", [Handle, ControlPid]),
		    megaco_udp_connect(MgcPort, RH, Handle, ControlPid);
		{error, Reason} ->
                    throw({error, {megaco_udp_open, Reason}})
	    end;
        {error, Reason} ->
            throw({error, {megaco_udp_start_transport, Reason}})
    end.

megaco_udp_connect(MgcPort, RH, Handle, ControlPid) ->
    {ok, LocalHost} = inet:gethostname(),
    MgcMid     = preliminary_mid,
    SendHandle = megaco_udp:create_send_handle(Handle, LocalHost, MgcPort),
    d("megaco connect", []),    
    {ok, CH}   = megaco:connect(RH, MgcMid, SendHandle, ControlPid),
    d("megaco connected: ~p", [CH]),    
    {ok, CH}.


do_apply_load(Parent, Mid) ->
    case ets:lookup(megaco_test_mg, Mid) of
	[MG] ->
	    [CH] = megaco:user_info(Mid, connections),
	    do_apply_load(Parent, Mid, CH, MG);
	[] ->
	    error
    end.

do_apply_load(Parent, Mid, CH, 
	      #mg{call_mode = Mode, group_size = Sz, load_counter = N0} = MG) ->
    d("do_apply_load -> entry with"
      "~n   Mode: ~p"
      "~n   Sz:   ~p"
      "~n   N0:   ~p", [Mode, Sz, N0]),
    {NofSent, Actions, ReplyData} = make_notify_request(N0, Sz),
    d("do_apply_load -> notifications constructed:"
      "~n   NofSent:   ~p"
      "~n   Actions:   ~p"
      "~n   ReplyData: ~p", [NofSent, Actions, ReplyData]),
    N = N0 - NofSent,
    case Mode of
	sync ->
	    Result = megaco:call(CH, Actions, []),
	    d("do_apply_load -> call result:~n   ~p", [Result]),
	    case N of
		0 ->
		    Parent ! {load_complete, Mid, self()},
		    ets:insert(megaco_test_mg, 
			       MG#mg{call_mode = async, load_counter = 0});
		_ ->
		    apply_load_timer(Mid),
		    ets:insert(megaco_test_mg, 
			       MG#mg{call_mode = async, load_counter = N})
	    end;
	async ->
	    Result = megaco:cast(CH, Actions, [{reply_data, ReplyData}]),
	    d("do_apply_load -> cast result:~n   ~p", [Result]),
	    ets:insert(megaco_test_mg, 
		       MG#mg{call_mode     = sync, 
			     load_counter  = N, 
			     reply_counter = NofSent}) % Outstanding replies
    end.


start_notify_request_handler(Mid, CH, N) ->
    d("start_notify_request_handler -> entry with"
      "~n   Mid: ~p"
      "~n   CH:  ~p"
      "~n   N:   ~p", [Mid, CH, N]),
    Env = get(),
    spawn_link(?MODULE, notify_request_handler_main, [self(), Env, Mid, CH, N]).

notify_request_handler_main(Parent, Env, Mid, CH, N) ->
    [put(Tag, Val) || {Tag, Val} <- Env],
    d("notify_request_handler_main -> entry with"
      "~n   Parent: ~p"
      "~n   CH:     ~p"
      "~n   N:      ~p", [Parent, CH, N]),
    Res = do_notify_request(CH, N),
    d("notify_request_handler_main -> notify complete:"
      "~n   Res: ~p", [Res]),
    Parent ! {notify_request_complete, Mid, {ok, Res}, self()},
    unlink(Parent),
    exit(normal).

do_notify_request(CH,N) when N =< 0 ->
    d("do_notify_request(~p) -> ignoring", [N]),
    ignore;
do_notify_request(CH,1) ->
    d("do_notify_request(1) -> entry with"),
    {Action, _} = make_notify_request(),
    megaco:call(CH, Action, []);
do_notify_request(CH,N) ->
    d("do_notify_request(~p) -> entry with", [N]),
    {N, Actions, _} = make_notify_request(N,N),
    megaco:call(CH, Actions, []).

make_notify_request(N, Sz) when N >= Sz, Sz > 0 ->
    {Req, ReplyData} = make_notify_request(N, Sz, [], []),
    {Sz, Req, ReplyData};
make_notify_request(N, _Sz) when N > 0 ->
    {Req, ReplyData} = make_notify_request(N, N, [], []),
    {N, Req, ReplyData}.


    
make_notify_request(_Offset, 0, Actions, ReplyDatas) ->
    {lists:reverse(Actions), lists:reverse(ReplyDatas)};
make_notify_request(Offset, N, Actions, ReplyDatas) when N > 0 ->
    TimeStamp = cre_timeNotation(),
    Event     = cre_observedEvent("al/of", TimeStamp),
    Desc      = cre_observedEventsDesc(2220 + N, [Event]),
    NotifyReq = cre_notifyReq([#megaco_term_id{id = tid(100+Offset-N)}],Desc),
    CmdReq    = cre_commandReq({notifyReq, NotifyReq}),
    ActReq    = cre_actionReq(?megaco_null_context_id, [CmdReq]),
    make_notify_request(Offset, N-1, [[ActReq]|Actions], [Desc|ReplyDatas]).
    
make_notify_request() ->
    TimeStamp  = cre_timeNotation("19990729", "22000000"),
    Event      = cre_observedEvent("al/of",   TimeStamp),
    Desc1      = cre_observedEventsDesc(2221, [Event]),
    Desc2      = cre_observedEventsDesc(2222, [Event]),
    Desc3      = cre_observedEventsDesc(2223, [Event]),
    Desc4      = cre_observedEventsDesc(2224, [Event]),
    NotifyReq1 = cre_notifyReq([#megaco_term_id{id = ?A4444}], Desc1),
    NotifyReq2 = cre_notifyReq([#megaco_term_id{id = ?A4445}], Desc2),
    CmdReq1    = cre_commandReq({notifyReq, NotifyReq1}),
    CmdReq2    = cre_commandReq({notifyReq, NotifyReq2}),
    ActReq     = cre_actionReq(?megaco_null_context_id, [CmdReq1,CmdReq2]),
    {[ActReq], [Desc1,Desc2]}.


cre_actionReq(Cid, Cmds) ->
    #'ActionRequest'{contextId       = Cid,
		     commandRequests = Cmds}.

cre_commandReq(Cmd) ->
    #'CommandRequest'{command = Cmd}.

cre_serviceChangeReq(TermId, Parms) ->
    #'ServiceChangeRequest'{terminationID      = TermId,
			    serviceChangeParms = Parms}.

cre_serviceChangeParm(Method, Reason) ->
    #'ServiceChangeParm'{serviceChangeMethod = Method,
			 serviceChangeReason = Reason}.

cre_notifyReq(Tid, EvsDesc) ->
    #'NotifyRequest'{terminationID = Tid, observedEventsDescriptor = EvsDesc}.

% cre_notifyRep(Tid) ->
%     #'NotifyReply'{terminationID = [Tid]}.

% cre_notifyRep(Tid,Err) ->
%     #'NotifyReply'{terminationID = [Tid], errorDescriptor = Err}.

cre_observedEventsDesc(Id, EvList) ->
    #'ObservedEventsDescriptor'{requestId = Id, observedEventLst = EvList}.

cre_observedEvent(Name, Not) ->
    #'ObservedEvent'{eventName = Name, timeNotation = Not}.

cre_timeNotation() ->
    {{Year,Month,Day},{Hour,Min,Sec}} = calendar:universal_time(),
    D = lists:flatten(io_lib:format("~4..0w~2..0w~2..0w", [Year,Month,Day])),
    T = lists:flatten(io_lib:format("~2..0w~2..0w~4..0w", [Hour,Min,Sec])),
    cre_timeNotation(D, T).
    
cre_timeNotation(D,T) ->
    #'TimeNotation'{date = D, time = T}.

cre_error_descr(Code,Text) ->
    #'ErrorDescriptor'{errorCode = Code, errorText = Text}.

% cre_error_descr(Code,FormatString,Args) ->
%     Text = lists:flatten(io_lib:format(FormatString,Args)),
%     cre_error_descr(Code,Text).


%% -----------------------
%% Handle megaco callbacks
%%

handle_megaco_request({handle_connect, _CH, _PV}, _Parent, Mid) ->
    d("handle_megaco_request(handle_connect) -> entry with"
      "~n   Mid: ~p", [Mid]),
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{state = connecting} = MG] ->
	    ok;
	[#mg{state = S} = MG] ->
	    Desc = lists:flatten(io_lib:format("not ready for connect in state ~p", [S])),
	    ED   =  cre_error_descr(?megaco_internal_gateway_error, Desc),
	    {discard_ack, ED};
	[] ->
	    d("handle_megaco_request(handle_connect) -> not found in:"
	      "~n   Tab: ~p", [ets:tab2list(megaco_test_mg)]),
	    Desc = lists:flatten(io_lib:format("unknown mg ~p", [Mid])),
	    ED   =  cre_error_descr(?megaco_internal_gateway_error, Desc),
	    {discard_ack, ED}
    end;

handle_megaco_request({handle_disconnect, _CH, _PV, R}, _Parent, Mid) ->
    d("handle_megaco_request(handle_disconnect) -> entry with"
      "~n   Mid: ~p", [Mid]),
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{req_handler = Pid} = MG] when pid(Pid) ->
	    d("handle_megaco_request(handle_disconnect) -> "
	      "~n   req_handler: ~p", [Pid]),
	    Error = {error, {disconnected, R}},
	    self() ! {notify_request_complete, Mid, Error, Pid},
	    unlink(Pid),
	    exit(Pid, kill),
	    MG1 = MG#mg{state = initiated},
	    ets:insert(megaco_test_mg, MG1),
	    ok;
	[MG] ->
	    d("handle_megaco_request(handle_disconnect) -> "
	      "~n   MG: ~p", [MG]),
	    ets:insert(megaco_test_mg, MG#mg{state = initiated}),

	    ok;
	[] ->
	    Desc = lists:flatten(io_lib:format("unknown mg ~p", [Mid])),
	    ED   =  cre_error_descr(?megaco_internal_gateway_error, Desc),
	    {discard_ack, ED}
    end;

handle_megaco_request({handle_syntax_error, _RH, _PV, _ED}, _Parent, _Mid) ->
    reply;

handle_megaco_request({handle_message_error, CH, PV, ED}, Parent, Mid) ->
    d("handle_megaco_request(handle_message_error) -> entry with"
      "~n   CH:     ~p"
      "~n   PV:     ~p"
      "~n   ED:     ~p"
      "~n   Parent: ~p"
      "~n   Mid:    ~p", [CH, PV, ED, Parent, Mid]),    
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{req_handler = Pid} = MG] when pid(Pid) ->
	    d("handle_megaco_request(handle_message_error) -> "
	      "~n   req_handler: ~p", [Pid]),
	    self() ! {notify_request_complete, Mid, ED, Pid},
	    unlink(Pid),
	    exit(Pid, kill),
	    ok;
	[] ->
	    ok
    end,
    no_reply;

handle_megaco_request({handle_trans_request, _CH, _PV, _AR}, _Parent, _Mid) ->
    ED =  cre_error_descr(?megaco_not_implemented,
			  "Transaction requests not handled"),
    {discard_ack, ED};

handle_megaco_request({handle_trans_long_request, _CH, _PV, _RD}, _Parent, _Mid) ->
    ED = cre_error_descr(?megaco_not_implemented,
			 "Long transaction requests not handled"),
    {discard_ack,  ED};

handle_megaco_request({handle_trans_reply, CH, PV, AR, RD}, Parent, Mid) ->
    do_handle_trans_reply(CH, PV, AR, RD, Parent, Mid);

handle_megaco_request({handle_trans_ack, _CH, _PV, _AS, _AD}, _Parent, _Mid) ->
    ok.

do_handle_trans_reply(CH, _PV, {ok, Rep}, _RD, Parent, Mid) ->
    d("do_handle_trans_reply -> entry with"
      "~n   CH:     ~p"
      "~n   Rep:    ~p"
      "~n   Parent: ~p"
      "~n   Mid:    ~p", [CH, Rep, Parent, Mid]),
    case ets:lookup(megaco_test_mg, Mid) of
	[#mg{state = connecting} = MG] ->
	    server_reply(Parent, service_change_reply, ok),
	    ets:insert(megaco_test_mg, MG#mg{state = connected}),
	    ok;
	[#mg{load_counter = 0} = MG] ->
	    d("do_handle_trans_reply(load_counter = 0) -> reply received ", []),
	    handle_trans_reply_verify_act(Rep),
	    server_reply(Parent, load_complete, ok),
	    ets:insert(megaco_test_mg, MG#mg{reply_counter = 0}),
	    ok;
	[#mg{reply_counter = 0} = MG] ->
	    d("do_handle_trans_reply(reply_counter = 0) -> reply received", []),
	    handle_trans_reply_verify_act(Rep),
	    apply_load_timer(Mid),
	    ok;
	[#mg{reply_counter = N} = MG] ->
	    d("do_handle_trans_reply(reply_counter = ~p) -> reply received ", [N]),
	    handle_trans_reply_verify_act(Rep),
	    ets:insert(megaco_test_mg, MG#mg{reply_counter = N-1}),
	    apply_load_timer(Mid),
	    ok;
	[] ->
	    Desc = lists:flatten(io_lib:format("unknown mg ~p", [Mid])),
	    ED   =  cre_error_descr(?megaco_internal_gateway_error, Desc),
	    {discard_ack, ED}
    end;
do_handle_trans_reply(_CH, _PV, {error, ED}, _RD, _Parent, _Mid) ->
    i("unexpected error transaction: ~p", [ED]),
    ok.


handle_trans_reply_verify_act([]) ->
    ok;
handle_trans_reply_verify_act([#'ActionReply'{commandReply = Rep}|Reps]) ->
    handle_trans_reply_verify_cmd(Rep),
    handle_trans_reply_verify_act(Reps);
handle_trans_reply_verify_act([Rep|Reps]) ->
    i("received 'propably' unexpected reply: ~n~p", [Rep]),
    handle_trans_reply_verify_act(Reps).

handle_trans_reply_verify_cmd([]) ->
    ok;
handle_trans_reply_verify_cmd([Cmd|Cmds]) ->
    case Cmd of
	{notifyReply, #'NotifyReply'{terminationID = [Tid]}} ->
	    d("received expected notification reply from ~n   ~p", [Tid]);
	Else ->
	    i("received unexpected notification reply ~n~p", [Else])
    end,
    handle_trans_reply_verify_cmd(Cmds).
    
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

notify_started(Parent) ->
    Parent ! {started, self()}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% The megaco user callback interface 

handle_connect(CH, PV, Pid, Mid) ->
    case CH#megaco_conn_handle.remote_mid of
        preliminary_mid ->
	    %% Avoids deadlock
	    ok;
	_ ->
	    Reply = request(Pid, {handle_connect, CH, PV}, Mid),
	    Reply
    end.

handle_disconnect(CH, PV, 
		  {user_disconnect, {stopped_by_user, Pid}}, 
		  Pid, Mid) ->
    ok;
handle_disconnect(CH, PV, R, Pid, Mid) ->
    request(Pid, {handle_disconnect, CH, PV, R}, Mid).

handle_syntax_error(ReceiveHandle, ProtocolVersion, ErrorDescriptor, Pid, Mid) ->
    Req = {handle_syntax_error, ReceiveHandle, ProtocolVersion, 
	   ErrorDescriptor},
    request(Pid, Req, Mid).
    
handle_message_error(ConnHandle, ProtocolVersion, ErrorDescriptor, Pid, Mid) ->
    Req = {handle_message_error, ConnHandle, ProtocolVersion, ErrorDescriptor},
    request(Pid, Req, Mid).

handle_trans_request(CH, PV, AR, Pid, Mid) ->
    Reply = request(Pid, {handle_trans_request, CH, PV, AR}, Mid),
    Reply.

handle_trans_long_request(ConnHandle, ProtocolVersion, ReqData, Pid, Mid) ->
    Req = {handle_trans_long_request, ConnHandle, ProtocolVersion, ReqData},
    request(Pid, Req, Mid).

handle_trans_reply(ConnHandle, ProtocolVersion, ActualReply, ReplyData, Pid, Mid) ->
    Req = {handle_trans_reply, ConnHandle, ProtocolVersion, 
	   ActualReply, ReplyData},
    request(Pid, Req, Mid).

handle_trans_ack(ConnHandle, ProtocolVersion, AckStatus, AckData, Pid, Mid) ->
    Req = {handle_trans_ack, ConnHandle, ProtocolVersion, AckStatus, AckData},
    request(Pid, Req, Mid).


request(Pid, Request, Mid) ->
    Pid ! {request, Request, Mid, self()},
    receive
	{reply, {delay, To, ED}, Pid} ->
	    sleep(To),
	    {discard_ack, ED};
	{reply, Reply, Pid} ->
	    Reply
    end.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sleep(X) ->
    receive after X -> ok end.


error_msg(F,A) -> error_logger:error_msg(F ++ "~n",A).


get_encoding_module(RI) ->
    case (catch get_conf(encoding_module, RI)) of
	{error, _} ->
	    undefined;
	Val ->
	    Val
    end.

get_encoding_config(RI, EM) ->
    case text_codec(EM) of
	true ->
	    case megaco:system_info(text_config) of
		[Conf] when list(Conf) ->
		    Conf;
		_ ->
		    []
	    end;

	false ->
	    get_conf(encoding_config, RI)
    end.

text_codec(megaco_compact_text_encoder) ->
    true;
text_codec(megaco_pretty_text_encoder) ->
    true;
text_codec(_) ->
    false.


get_transport_module(RI) ->
    get_conf(transport_module, RI).

get_transport_port(RI) ->
    get_conf(port, RI).


get_conf(Key, Config) ->
    case lists:keysearch(Key, 1, Config) of
	{value, {Key, Val}} ->
	    Val;
	_ ->
	    exit({error, {not_found, Key, Config}})
    end.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tid(N) when N >= 0 ->
    {Rem1, Val1} = num2str(N),
    {Rem2, Val2} = num2str(Rem1),
    {0,    Val3} = num2str(Rem2),
    [Val3, Val2, Val1].

num2str(N) when N >= 0 ->
    num2str(N, []).

num2str(Rem, Val) when length(Val) == 8 ->
    {Rem, Val};
num2str(N, Val) ->
    D = N div 2,
    case N rem 2 of
	1 ->
	    num2str(D, [$1|Val]);
	0 -> 
	    num2str(D, [$0|Val])
    end.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

i(F) ->
    i(F, []).

i(F, A) ->
    print(info, get(verbosity), "", F, A).


d(F) ->
    d(F, []).

d(F, A) ->
    print(debug, get(verbosity), "DBG: ", F, A).


printable(_, debug)   -> true;
printable(info, info) -> true;
printable(_,_)        -> false.

print(Severity, Verbosity, P, F, A) ->
    print(printable(Severity,Verbosity), P, F, A).

print(true, P, F, A) ->
    io:format("*** [~s] ~s ~p ~s ***"
              "~n   " ++ F ++ "~n~n", 
              [format_timestamp(now()), P, self(), get(sname) | A]);
print(_, _, _, _) ->
    ok.

format_timestamp(Now) ->
    {N1, N2, N3} = Now,
    {Date, Time}   = calendar:now_to_datetime(Now),
    {YYYY,MM,DD}   = Date,
    {Hour,Min,Sec} = Time,
    FormatDate = 
        io_lib:format("~.4w:~.2.0w:~.2.0w ~.2.0w:~.2.0w:~.2.0w 4~w",
                      [YYYY,MM,DD,Hour,Min,Sec,round(N3/1000)]),  
    lists:flatten(FormatDate).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

random_init() ->
    {A,B,C} = now(),
    random:seed(A,B,C).

random() ->
    10 * random:uniform(50).

apply_load_timer(Mid) ->
    erlang:send_after(random(), self(), {apply_load_timeout, Mid}).

